#!/usr/bin/env python

from __future__ import print_function
import numpy
import vtk


class Triangulation:

    def __init__(self):
        """
        Constructor
        """
        self.points = vtk.vtkPoints()
        self.polydata = vtk.vtkPolyData()
        self.delny = vtk.vtkDelaunay3D()

        self.polydata.SetPoints(self.points)
        if vtk.VTK_MAJOR_VERSION >= 6:
            self.delny.SetInputData(self.polydata)
        else:
            self.delny.SetInput(self.polydata)

    def setInputPoints(self, points,):
        """
        Set input points
        @param points points
        """
        numPoints = points.shape[0]
        for i in range(numPoints):
            self.points.InsertPoint(i, points[i, 0],
                                    points[i, 1], points[i, 2])

    def triangulate(self):
        """
        Triangulate
        """
        self.delny.Update()

    def getVTKUnstructuredGrid(self):
        """
        Get the VTK unstructured grid
        @return VTK object
        """
        return self.delny.GetOutput()

    def getCells(self):
        """
        Get cells
        @return list of vertex indices
        """
        cells = self.delny.GetOutput().GetCells()
        numCells = cells.GetNumberOfCells()

        npts = 4  # number of points per cell
        cellArr = numpy.zeros((numCells, npts), numpy.int)
        ptIds = vtk.vtkIdList()
        for i in range(numCells):
            cells.GetCell(i, ptIds)
            for j in range(npts):
                cellArr[i, j] = int(ptIds.GetId(j))

        return cellArr

##############################################################################


def test():

    points = numpy.array([(0, 0, 0), (0, 0, 1), (0, 1, 0),
                          (0, 1, 1), (1, 0, 0), (1, 0, 1),
                          (1, 1, 0), (1, 1, 1)], numpy.float64)
    tri = Triangulation()
    tri.setInputPoints(points)
    tri.triangulate()
    print(tri.getVTKUnstructuredGrid())
    cells = tri.getCells()
    print('cells = ', cells)

if __name__ == '__main__':
    test()



#!/usr/bin/env python

from __future__ import print_function
import math
import numpy
import vtk
import triangle


class RefineSurface:

    def __init__(self, pdata):
        """
        Constructor
        @param pdata vtkPolyData instance
        """
        self.polydata = pdata

        # save the points and point data as separate class members
        # so as not to pollute pdata
        self.points = vtk.vtkPoints()
        self.points.DeepCopy(pdata.GetPoints())

        self.pointData = {}
        pd = pdata.GetPointData()
        for i in range(pd.GetNumberOfArrays()):
            arr = pd.GetArray(i)
            name = arr.GetName()
            self.pointData[name] = vtk.vtkDoubleArray()
            self.pointData[name].DeepCopy(arr)

        self.cellData = {}
        cd = pdata.GetCellData()
        for i in range(cd.GetNumberOfArrays()):
            name = cd.GetArray(i).GetName()
            self.cellData[name] = vtk.vtkDoubleArray()
            self.cellData[name].SetName(name)

    def getVtkPolyData(self):
        """
        Get the vtkPolyData instance associated with the grid
        @return vtkPolyData instance
        """
        return self.polydata

    def refine(self, max_edge_length):
        """
        Refine each cell by adding points on edges longer than max_edge_length
        @param max_edge_length maximum edge length (> 0)
        @note operation is in place
        """

        # no operation if max_edge_length is not strictly positive
        if max_edge_length <= 0:
            return

        # iterate over the polys
        polys = self.polydata.GetPolys()
        ptIds = vtk.vtkIdList()
        polys.InitTraversal()
        cells = []
        edge2PtIds = {}
        newCellBegEndInds = []
        iNewBeg = 0
        for iPoly in range(polys.GetNumberOfCells()):

            polys.GetNextCell(ptIds)
            numPolyPts = ptIds.GetNumberOfIds()
            polyPtIds = []

            if numPolyPts < 3:
                # need at least three points, next polygon
                continue

            # compute the two tangential unit vectors
            uVec, vVec, normal = self.computeUVNormal(ptIds)
            if normal.dot(normal) == 0:
                # zero area polygon, nothing to do
                continue

            # iterate over edges
            for iNode in range(numPolyPts):
                i0 = ptIds.GetId(iNode)
                i1 = ptIds.GetId((iNode + 1) % numPolyPts)
                edge = (i0, i1)
                edgeCompl = (i1, i0)

                edgePtIds = edge2PtIds.get(edgeCompl, [])
                if len(edgePtIds) > 0:
                    # edge has already been split, take edge and reverse order
                    edgePtIds.reverse()
                    edgePtIds = [i0] + edgePtIds[:-1]
                    polyPtIds += edgePtIds
                    continue

                p0 = numpy.array(self.points.GetPoint(i0))
                p1 = numpy.array(self.points.GetPoint(i1))
                dEdge = p1 - p0
                edgeLength = numpy.sqrt(dEdge.dot(dEdge))
                numSegs = int(max(1, math.ceil(edgeLength/max_edge_length)))

                # add points along edge
                d10 = (p1 - p0) / float(numSegs)
                edgePtIds.append(i0)
                for iSeg in range(1, numSegs):
                    pt = p0 + iSeg*d10
                    # index of point to add
                    ptId = self.points.GetNumberOfPoints()
                    # insert point
                    self.points.InsertNextPoint(pt)
                    edgePtIds.append(ptId)
                    # interpolate the field values along the edge
                    w0 = (numSegs - iSeg)/float(numSegs)
                    w1 = iSeg/float(numSegs)
                    for name in self.pointData:
                        v0 = numpy.array(self.pointData[name].GetTuple(i0))
                        v1 = numpy.array(self.pointData[name].GetTuple(i1))
                        interpTuple = w0*v0 + w1*v1
                        self.pointData[name].InsertNextTuple(interpTuple)

                edge2PtIds[edge] = edgePtIds
                polyPtIds += edgePtIds

            # triangulate the cell
            polyCells = self.triangulatePolygon(uVec, vVec, polyPtIds,
                                                max_edge_length)
            cells += polyCells
            iNewEnd = iNewBeg + len(polyCells)
            # all the cell data have the same value inside this polygon
            # store the srta/end cell indices 
            newCellBegEndInds.append( (iNewBeg, iNewEnd) )
            iNewBeg = iNewEnd

        # build the output vtkPolyData object
        newPolyData = vtk.vtkPolyData()

        # set points
        newPolyData.SetPoints(self.points)

        # set all the point data
        for name in self.pointData:
            newPolyData.GetPointData().AddArray(self.pointData[name])

        # build the connectivity
        numPolys = len(cells)
        newPolyData.Allocate(numPolys, 1)
        ptIds = vtk.vtkIdList()
        for cell in cells:
            numPolyPts = len(cell)
            ptIds.SetNumberOfIds(numPolyPts)
            for j in range(numPolyPts):
                ptIds.SetId(j, cell[j])
            newPolyData.InsertNextCell(vtk.VTK_POLYGON, ptIds)

        numTuples = self.polydata.GetPolys().GetNumberOfCells()
        newNumTuples = newPolyData.GetPolys().GetNumberOfCells()

        # add the cell arrays
        cd = self.polydata.GetCellData()
        numCellArrays = cd.GetNumberOfArrays()
        for el in range(numCellArrays):
            arr = cd.GetArray(el)
            name = arr.GetName()
            numComps = arr.GetNumberOfComponents()

            newArr = self.cellData[name]
            newArr.SetNumberOfComponents(numComps)
            newArr.SetNumberOfTuples(newNumTuples)
            for i in range(numTuples):
                vals = arr.GetTuple(i)
                iNewBeg, iNewEnd = newCellBegEndInds[i]
                for j in range(iNewBeg, iNewEnd):
                    newArr.SetTuple(j, vals)
            newPolyData.GetCellData().AddArray(newArr)

        # Reset the polydata struct
        self.polydata = newPolyData

    def computeUVNormal(self, ptIds):
        """
        Compute the two tangential unit vectors and the normal vector
        @param ptIds point indices
        @return u vector, v vector, normal
        """
        uVec = numpy.zeros((3,), numpy.float64)
        vVec = numpy.zeros((3,), numpy.float64)
        normal = numpy.zeros((3,), numpy.float64)
        numPts = ptIds.GetNumberOfIds()
        if numPts < 3:
            return uVec, vVec, normal
        p0 = numpy.array(self.points.GetPoint(ptIds.GetId(0)))
        for i in range(1, numPts - 1):
            dp1 = numpy.array(self.points.GetPoint(ptIds.GetId(i))) - p0
            dp2 = numpy.array(self.points.GetPoint(ptIds.GetId(i + 1))) - p0
            perp = numpy.cross(dp1, dp2)
            pDotp = perp.dot(perp)
            if pDotp > 0:
                normal = perp / numpy.sqrt(pDotp)
                uVec = dp1 / numpy.sqrt(dp1.dot(dp1))
                vVec = numpy.cross(normal, uVec)
                return uVec, vVec, normal
        return uVec, vVec, normal

    def triangulatePolygon(self, uVec, vVec, polyPtIds, max_edge_length):
        """
        Triangulate polygon using the uVec x vVec projection
        @param uVec unit vector tangential to the polygon (input)
        @param vVec second unit vector tangential to the polygon (input)
        @param polyPtIds list of polygon's point indices (input and output)
        @param max_edge_length maximum edge length (input)
        @return list of cells (list of point indices)
        """

        numPolyPts = len(polyPtIds)
        if numPolyPts < 3:
            # need at least three points
            return []

        area = self.getPolygonArea(polyPtIds)
        if area < 1.e-15:
            return []

        self.removeDegeneratePoints(polyPtIds)

        # reset
        numPolyPts = len(polyPtIds)
        if numPolyPts < 3:
            return []

        # reference position
        p0 = numpy.array(self.points.GetPoint(polyPtIds[0]))

        pts = []
        pointData = {}
        for name in self.pointData:
            pointData[name] = []
        # project each point onto the plane
        for i in range(numPolyPts):
            ptId = polyPtIds[i]
            pos = numpy.array(self.points.GetPoint(ptId))
            pos -= p0
            pts.append((pos.dot(uVec), pos.dot(vVec)))
            for name, pd in pointData.items():
                pd.append(tuple(self.pointData[name].GetTuple(ptId)))

        # list of segments
        segs = [(i, (i + 1)%numPolyPts) for i in range(numPolyPts)]

        tri = triangle.Triangle()
        tri.set_points(pts)
        tri.set_segments(segs)
        attrs, names, components = self.pointDataToAttributes(pointData)
        tri.set_attributes(attrs)

        # internal points will be added if triangle area exceeds threshold
        maxArea = None
        if max_edge_length < float('inf') and max_edge_length > 0.:
            maxArea = 0.5 * max_edge_length**2

        # triangulate
        # p: triangulate a straight planar graph
        # z: zero based indexing
        # Q: quiet mode
        tri.triangulate(area=maxArea, mode='pzQ')

        nodes = tri.get_nodes()
        polyCells = tri.get_triangles()
        interpolatedAttrs = tri.get_attributes()

        # add internal vertices
        pIndex2PtId = {}
        for pIndex in range(len(pts), len(nodes)):
            ptId = self.points.GetNumberOfPoints()
            u, v = nodes[pIndex][0]
            p = p0 + u*uVec + v*vVec
            # insert new point
            self.points.InsertNextPoint(p)
            polyPtIds.append(ptId)
            pIndex2PtId[pIndex] = ptId

        # add point data inside the cell
        for name, pd in self.pointData.items():
            numTuples = pd.GetNumberOfTuples()
            newSize = numTuples + len(nodes) - len(pts)
            success = pd.Resize(newSize)
            if success != 1:
                msg = 'Failed to resize vtkDoubleArray associated with field {0}'.format(name)
                raise MemoryError(msg)
            pd.SetNumberOfTuples(newSize)
            pd.SetName(name)

            # add the new internal point data
            iBeg = names.index(name)
            iEnd = len(names) - names[::-1].index(name)
            for pIndex in range(len(pts), len(nodes)):
                ptId = pIndex2PtId[pIndex]
                vals = interpolatedAttrs[pIndex][iBeg:iEnd]
                pd.SetTuple(ptId, vals)

        cells = []
        for c in polyCells:
            ia, ib, ic = c[0]
            cell = [polyPtIds[ia], polyPtIds[ib], polyPtIds[ic]]
            cells.append(cell)

        return cells

    def getPolygonArea(self, polyPtIds):
        """
        Compute the (scalar) area of a polygon
        @param polyPtIds list of point indices
        @return area
        """
        area = numpy.zeros((3,), numpy.float64)
        p0 = numpy.array(self.points.GetPoint(polyPtIds[0]))
        numPolyPts = len(polyPtIds)
        for i in range(1, numPolyPts - 1):
            p1 = numpy.array(self.points.GetPoint(polyPtIds[i]))
            p2 = numpy.array(self.points.GetPoint(polyPtIds[i + 1]))
            area += numpy.cross(p1 - p0, p2 - p0)
        return numpy.linalg.norm(area)

    def removeDegeneratePoints(self, polyPtIds):
        """
        Remove degenerate points
        @param polyPtIds list of point indices (modified on output)
        """
        indicesToDelete = []
        numPolyPts = len(polyPtIds)
        for i in range(numPolyPts):
            p0 = numpy.array(self.points.GetPoint(polyPtIds[i]))
            p1 = numpy.array(self.points.GetPoint(polyPtIds[(i + 1)%numPolyPts]))
            if numpy.linalg.norm(p1 - p0) < 1.e-15:
                indicesToDelete.append(i)
        for j in range(len(indicesToDelete) - 1, -1, -1):
            i = indicesToDelete[j]
            del polyPtIds[i]
        
    def pointDataToAttributes(self, pointData):
        """
        Convert point data dictionary 
        into an attribute array which can be passed to the triangle program
        @param pointData {name: [(comp0, comp1, ...), ...], ...}
        @return attributes, names, and components as separate lists
        """
        attrs = []
        names = []
        components = []

        if len(pointData) == 0:
            return attrs, names, components

        scalarNames = list(pointData.keys())
        scalarData = list(pointData.values())
        
        # set the name and component lists, which is 
        # the same for each node
        for j in range(len(scalarNames)):
            name = scalarNames[j]
            data = scalarData[j]
            for k in range(len(data[0])):
                names.append(name)
                components.append(k)
                
        numPts = len(scalarData[0])
        for i in range(numPts):
            a = []

            # iterate ove the scalar field names
            for j in range(len(scalarNames)):
                data = scalarData[j]
          
                # iterate over the components
                for k in range(len(data[0])):
                    a.append(data[i][k])

            attrs.append(tuple(a))

        return attrs, names, components

##############################################################################


def printVtkPolyData(pdata):

    points = pdata.GetPoints()
    polys = pdata.GetPolys()
    ptIds = vtk.vtkIdList()
    numPolys = polys.GetNumberOfCells()
    print('Number of polygons: {0}'.format(numPolys))
    polys.InitTraversal()
    for i in range(numPolys):
        numPts = ptIds.GetNumberOfIds()
        print('\tCell {0} has {1} points: '.format(i, numPts))
        for j in range(numPts):
            ptId = ptIds.GetId(j)
            pt = points.GetPoint(ptId)
            print('\t\t{0} -> {1}'.format(ptId, pt))


def testNoRefinement():

    points = vtk.vtkPoints()
    points.InsertNextPoint((0., 0., 0.))
    points.InsertNextPoint((1., 0., 0.))
    points.InsertNextPoint((1., 1., 0.))

    pdata = vtk.vtkPolyData()
    pdata.SetPoints(points)

    ptIds = vtk.vtkIdList()
    ptIds.InsertNextId(0)
    ptIds.InsertNextId(1)
    ptIds.InsertNextId(2)
    pdata.Allocate(1, 1)
    pdata.InsertNextCell(vtk.VTK_POLYGON, ptIds)

    rs = RefineSurface(pdata)
    rs.refine(max_edge_length=1.5)
    pdata2 = rs.getVtkPolyData()
    assert(pdata2.GetNumberOfPolys() == 1)


def testAddingThreePointsThenMore():

    points = vtk.vtkPoints()
    points.InsertNextPoint((0., 0., 0.))
    points.InsertNextPoint((2., 0., 0.))
    points.InsertNextPoint((2., 1., 0.))

    pdata = vtk.vtkPolyData()
    pdata.SetPoints(points)

    ptIds = vtk.vtkIdList()
    ptIds.InsertNextId(0)
    ptIds.InsertNextId(1)
    ptIds.InsertNextId(2)
    pdata.Allocate(1, 1)
    pdata.InsertNextCell(vtk.VTK_POLYGON, ptIds)

    rs = RefineSurface(pdata)
    rs.refine(max_edge_length=1.1)
    assert(rs.getVtkPolyData().GetNumberOfPolys() == 4)


def testStartingWithTwoCells():

    points = vtk.vtkPoints()
    points.InsertNextPoint((0., 0., 0.))
    points.InsertNextPoint((2., 0., 0.))
    points.InsertNextPoint((2., 1., 0.))
    points.InsertNextPoint((0., 1., 0.))

    pdata = vtk.vtkPolyData()
    pdata.SetPoints(points)

    pdata.Allocate(2, 1)
    ptIds = vtk.vtkIdList()
    ptIds.SetNumberOfIds(3)
    ptIds.SetId(0, 0)
    ptIds.SetId(1, 1)
    ptIds.SetId(2, 2)
    pdata.InsertNextCell(vtk.VTK_POLYGON, ptIds)
    ptIds.SetId(0, 2)
    ptIds.SetId(1, 3)
    ptIds.SetId(2, 0)
    pdata.InsertNextCell(vtk.VTK_POLYGON, ptIds)

    rs = RefineSurface(pdata)
    rs.refine(max_edge_length=1.1)
    assert(rs.getVtkPolyData().GetNumberOfPolys() == 8)


if __name__ == '__main__':
    testNoRefinement()
    testAddingThreePointsThenMore()
    testStartingWithTwoCells()



#!/usr/bin/env python

from __future__ import print_function
import vtk
import numpy
from icqsol.bem.icqBaseLaplaceSolver import BaseLaplaceSolver


class PoissonSolver(BaseLaplaceSolver):

    def __init__(self, pdata, max_edge_length, order=5):
        """
        Constructor
        @param pdata instance of vtkPolyData
        @param max_edge_length maximum edge length, used to turn
                               polygons into triangles
        """
        #super(BaseLaplaceSolver, self).__init__(pdata, max_edge_length, order)
        BaseLaplaceSolver.__init__(self, pdata, max_edge_length, order)
        self.responseName = 'v'
        self.sourceName = 'charge'

    def computeResponseField(self):
        """
        Compute the response field, in this case the potential due to a charge source
        @return response
        """
        
        srcIndex = self.getSourceArrayIndex()
        src = self.getSourceArray(srcIndex)

        # Get the response matrix.
        gMat = self.getGreenMatrix()

        # Compute the response.
        rsp = numpy.dot(gMat, src)

        self.addResponseField(rsp)

        return rsp

###############################################################################


def testSingleTriangle():

    "Single triangle"

    h = 0.1
    # create set of points
    points = vtk.vtkPoints()
    points.SetNumberOfPoints(3)
    points.SetPoint(0, [1., -1.*h/3., -1.*h/3.])
    points.SetPoint(1, [1., 2.*h/3., -1.*h/3.])
    points.SetPoint(2, [1., -1.*h/3., 2.*h/3.])

    # create vtkPolyData object
    pdata = vtk.vtkPolyData()
    pdata.SetPoints(points)
    ptIds = vtk.vtkIdList()
    ptIds.SetNumberOfIds(3)
    ptIds.SetId(0, 0)
    ptIds.SetId(1, 1)
    ptIds.SetId(2, 2)
    pdata.Allocate(1, 1)
    pdata.InsertNextCell(vtk.VTK_POLYGON, ptIds)

    for order in range(1, 6):
        lslm = PoissonSolver(pdata, max_edge_length=1000.)
        print('order = ', order)
        print('g matrix: ', lslm.getGreenMatrix())


def testTwoTrianglesCoplanar():

    "Two triangles"

    # create set of points
    points = vtk.vtkPoints()
    points.SetNumberOfPoints(4)
    points.SetPoint(0, [0., 0., 0.])
    points.SetPoint(1, [1., 0., 0.])
    points.SetPoint(2, [0., 1., 0.])
    points.SetPoint(3, [1., 1., 0.])

    # create vtkPolyData object
    pdata = vtk.vtkPolyData()
    pdata.SetPoints(points)

    pdata.Allocate(2, 1)
    ptIds = vtk.vtkIdList()
    ptIds.SetNumberOfIds(3)

    ptIds.SetId(0, 0)
    ptIds.SetId(1, 1)
    ptIds.SetId(2, 2)
    pdata.InsertNextCell(vtk.VTK_POLYGON, ptIds)
    ptIds.SetId(0, 1)
    ptIds.SetId(1, 3)
    ptIds.SetId(2, 2)
    pdata.InsertNextCell(vtk.VTK_POLYGON, ptIds)

    for order in range(1, 6):
        lslm = PoissonSolver(pdata,
                             max_edge_length=1000.,
                             order=order)
        print('order = ', order)
        print('g matrix: ', lslm.getGreenMatrix())


def testTwoTriangles():

    "Two triangles"

    # create set of points
    points = vtk.vtkPoints()
    points.SetNumberOfPoints(4)
    points.SetPoint(0, [0., 0., 0.])
    points.SetPoint(1, [1., 0., 0.])
    points.SetPoint(2, [0., 1., 0.])
    points.SetPoint(3, [0., 0., 1.])

    # create vtkPolyData object
    pdata = vtk.vtkPolyData()
    pdata.SetPoints(points)

    pdata.Allocate(2, 1)
    ptIds = vtk.vtkIdList()
    ptIds.SetNumberOfIds(3)

    ptIds.SetId(0, 0)
    ptIds.SetId(1, 1)
    ptIds.SetId(2, 3)
    pdata.InsertNextCell(vtk.VTK_POLYGON, ptIds)
    ptIds.SetId(0, 0)
    ptIds.SetId(1, 3)
    ptIds.SetId(2, 2)
    pdata.InsertNextCell(vtk.VTK_POLYGON, ptIds)

    for order in range(1, 6):
        lslm = PoissonSolver(pdata,
                             max_edge_length=1000.,
                             order=order)
        print('order = ', order)
        print('g matrix: ', lslm.getGreenMatrix())

if __name__ == '__main__':
    testSingleTriangle()
    testTwoTrianglesCoplanar()
    testTwoTriangles()



#!/usr/bin/env python
"""
Map an arbitrary triangle to a reference triangle
http://arxiv.org/pdf/1201.4938.pdf
"""

from __future__ import print_function
import numpy
import math


class ReferenceTriangle:

    def __init__(self, xa, xb, xc):
        """
        Constructor
        @param xa first triangle vertex
        @param xb second triangle vertex
        @param xc third triangle vertex
        """

        # triangle edge lengths
        x0 = numpy.array(xa)
        self.xbdiff = numpy.array(xb) - x0
        self.xcdiff = numpy.array(xc) - x0

        xbdiffCrossXcdiff = numpy.cross(self.xbdiff, self.xcdiff)

        # the normal vector at the reference position
        self.normal = xbdiffCrossXcdiff / numpy.linalg.norm(xbdiffCrossXcdiff)

        # the edge length at theta = 0
        self.r1 = numpy.linalg.norm(self.xbdiff)

        # the edge length at theta = bigTheta
        self.r2 = numpy.linalg.norm(self.xcdiff)

        # the max integration angle
        self.bigTheta = math.atan2(
            numpy.linalg.norm(xbdiffCrossXcdiff), self.xbdiff.dot(self.xcdiff)
            )

        # the slope of the b-c triangle edge, x = r1 + a*y
        self.a = (self.r2*math.cos(self.bigTheta) - self.r1) / \
            self.r2/math.sin(self.bigTheta)

    def getPointBDiff(self):
        """
        Get the difference between the second and the first vertex points
        @return vector
        """
        return self.xbdiff

    def getPointCDiff(self):
        """
        Get the difference between the third and the first vertex points
        @return vector
        """
        return self.xcdiff

    def getR1(self):
        """
        Get the triangle length "r1" along theta = 0
        @return length
        """
        return self.r1

    def getR2(self):
        """
        Get the triangle length "r2" along theta = bigTheta
        @return length
        """
        return self.r2

    def getA(self):
        """
        Get the coefficient "a" in the opposite edge equation x' = r1 + a*y'
        @return coefficient
        """
        return self.a

    def getBigTheta(self):
        """
        Get the max theta angle of integration
        @return angle in radians
        """
        return self.bigTheta

##########################################################################


def test1():
    # simple, right-angle triangle
    rt = ReferenceTriangle([0., 0., 0.], [1., 0., 0.], [0., 2., 0.])
    assert(abs(rt.getR1() - 1.0) < 1.e-10)
    assert(abs(rt.getR2() - 2.0) < 1.e-10)
    assert(abs(rt.getA() + 1.0/2.0) < 1.e-10)
    assert(abs(rt.getBigTheta() - math.pi/2.0) < 1.e-10)


def test2():
    # triangle with offset
    rt = ReferenceTriangle([10., 20., 30.], [11., 20., 30.], [10., 22., 30.])
    assert(abs(rt.getR1() - 1.0) < 1.e-10)
    assert(abs(rt.getR2() - 2.0) < 1.e-10)
    assert(abs(rt.getA() + 1.0/2.0) < 1.e-10)
    assert(abs(rt.getBigTheta() - math.pi/2.0) < 1.e-10)


def test3():
    # big angle
    rt = ReferenceTriangle([10., 20., 30.], [11., 20., 30.], [9.0, 22., 30.])
    assert(abs(rt.getR1() - 1.0) < 1.e-10)
    assert(abs(rt.getR2() - math.sqrt(5.)) < 1.e-10)
    assert(abs(rt.getA() + 1.0) < 1.e-10)

if __name__ == '__main__':
    test1()
    test2()
    test3()



#!/usr/bin/env python

from __future__ import print_function
import numpy

class FiniteElementIntegrals:

	def __init__(self, vertices):
		"""
		Constructor
		@param vertices list of vertices
		"""

		# distance along each edge of the tettrahedron's vertices
		rEdge = [numpy.array(vertices[1]) - numpy.array(vertices[0]), 
		         numpy.array(vertices[2]) - numpy.array(vertices[0]), 
		         numpy.array(vertices[3]) - numpy.array(vertices[0])]

		# the parallelepiped volume of the cell
		self.jac = numpy.dot(numpy.cross(rEdge[0], rEdge[1]), rEdge[2])

		# an approximation of the co-variant metric tensor
		dxMat = numpy.zeros( (3, 3), numpy.float64 )
		for i in range(3):
			for j in range(3):
				dxMat[i, j] = numpy.dot(rEdge[i], rEdge[j])

		# contravariant approximation of the metric tensor
		self.invDxMat = numpy.linalg.inv(dxMat)

		self.dBasisOverDXi = numpy.array( [[-1., -1., -1.], 
			                               [1., 0., 0.], 
			                               [0., 1., 0.], 
			                               [0., 0., 1.]])

		# derivative of basis function with respect to cartesian coordinates
		self.grads = numpy.dot(self.dBasisOverDXi, self.invDxMat)

	def integrateBasisBasis(self):
		"""
		Compute the basis function times basis function cell integral
		@return a 4x4, contribution matrix containing the coupling between nodes
		"""
		res = numpy.zeros( (4, 4), numpy.float64)
		for i in range(4):
			res[i, i] = self.jac / 60.0
			res[i, i+1:] = self.jac / 120.0
			res[i+1:, i] = self.jac / 120.0
		return res

	def integrateGradientDotGradient(self):
		"""
		Compute the cell integral of the gradient of the basis function 
		dotted with the gradient of another gradient basis function
		@return a 4x4, contribution matrix containing the coupling between nodes
		"""
		integral = 1. / 6.
		res = numpy.zeros( (4, 4), numpy.float64 )
		# iterate over basis functions
		for iBasis1 in range(4):
			for iBasis2 in range(iBasis1, 4):
				dotProd = numpy.dot(self.grads[iBasis1, :], 
					                self.grads[iBasis2, :])
				res[iBasis1, iBasis2] = self.jac * integral * dotProd
				res[iBasis2, iBasis1] = res[iBasis1, iBasis2]
		return res

##############################################################################
def test():
	v0 = (0., 0., 0.)
	v1 = (1., 0., 0.)
	v2 = (0., 1., 0.)
	v3 = (0., 0., 1.)
	verts = (v0, v1, v2, v3)
	fei = FiniteElementIntegrals(verts)
	print('basis basis = ', fei.integrateBasisBasis())
	print('grad dot grad = ', fei.integrateGradientDotGradient())

if __name__ == '__main__':
	test()



#!/usr/bin/env python

"""
@brief A class for testing if a point is inside a shape
@author alexander@gokliya.net
"""

from __future__ import print_function
import numpy
import math
from functools import reduce


class Inside:

    def __init__(self, shape):
        """
        Constructor
        @param shape instance of Shape
        @param polys connectivty
        """
        points, polys, count = shape.toVerticesAndPolygons()
        # must have at least one point
        self.ndims = len(points[0])

        # find the box corners
        self.xmins = numpy.array([min([p[i] for p in points]) for
                                 i in range(self.ndims)])
        self.xmaxs = numpy.array([max([p[i] for p in points]) for
                                 i in range(self.ndims)])

        self.eps = 1.23456789e-14

        self.polys = polys
        self.points = [numpy.array(p) for p in points]

        # matrix and right and side vector
        self.mat = numpy.zeros((self.ndims, self.ndims), numpy.float64)
        self.b = numpy.zeros((self.ndims,), numpy.float64)

        # the optimal direction for shooting the ray
        self.direction = float('inf') * \
            numpy.ones((self.ndims,), numpy.float64)

        # min/max box corners of the ray
        self.xminRay = float('inf') * numpy.ones((self.ndims,), numpy.float64)
        self.xmaxRay = float('inf') * numpy.ones((self.ndims,), numpy.float64)

    def isInside(self, point, minDistance):
        """
        Determine if a point is inside the shape
        @param point point
        @param minDistance a point is declared inside if is at least minDistance
               away from the face
        @return +1 if inside, -1 if outside, and 0 if indefinite
        """

        # quick check, point must be inside box
        if reduce(lambda x, y: x or y,
                  [(point[i] < self.xmins[i] - minDistance) or
                   (point[i] > self.xmaxs[i] + minDistance)
                   for i in range(self.ndims)]):
            # outside box
            return -1

        # any direction will do but things will run faster if the direction
        # points to the nearest domain box face (fewer intersections to
        #  compute)
        self.computeOptimalDirection(point)

        # set the first column in our matrix (independent of the poly)
        self.mat[:, 0] = self.direction

        # compute the number of intersections between the ray and the polys
        numberOfIntersections = 0
        for poly in self.polys:

            isTriangle = 1
            numPoints = len(poly)
            if numPoints == 4:
                isTriangle = 0

            # compute the overlap betwen the ray and face boxes
            if not self.areBoxesOverlapping(point, poly, minDistance):
                # intersection is impossible, don't bother...
                continue

            lmbda, xis = self.computeIntersection(point, poly)

            if lmbda > -self.eps:

                # the direction is ok

                sums = 0.0
                rayIntersects = True
                for i in range(len(xis)):
                    rayIntersects &= (xis[i] >= 0.0)
                    rayIntersects &= (xis[i] < 1.0 - isTriangle*sums)
                    sums += xis[i]

                if rayIntersects:

                    target = lmbda * self.direction
                    distancePointToFace = math.sqrt(numpy.dot(target, target))
                    if distancePointToFace <= minDistance:
                        # marginal, cannot say
                        return 0
                    else:
                        numberOfIntersections += 1

        # even number is outside, odd number means inside
        return 2*(numberOfIntersections % 2) - 1

    def areBoxesOverlapping(self, point, poly, minDistance):
        """
        Determine if there is overlap between the ray and the poly min/max
        coordinates
        @param point starting point of the ray
        @param poly the face
        @param minDistance the distance below which we cannot say for sure
               whether there is an overlap or not
        """

        for i in range(self.ndims):

            xminFace = min([self.points[j][i] for j in poly])
            xmaxFace = max([self.points[j][i] for j in poly])

            if (self.xmaxRay[i] < xminFace - minDistance) or \
               (self.xminRay[i] > xmaxFace + minDistance):
                # no overlap possible
                return False
        # there is a strong possibility of overlap
        return True

    def computeOptimalDirection(self, point):
        """
        Compute the direction of the ray to the nearest
        domain box face and perturb that direction slightly 
        to minimize the probability that the ray hits a  node
        @param point starting point of the ray
        @note this will update self.direction
        """
        # iterate over the faces of the box then find the minimum
        # distance between the point and the face.
        minDistance = float('inf')

        # high/low side of the box
        for pm in (-1, 1):
            pls = (1 + pm)/2.  # 0 on the low side, 1 on the high side
            mns = (1 - pm)/2.  # 1 on the low side, 0 on the high side

            # iterate over the axes
            for axis in range(self.ndims):
                # the normal vector contains very small values in place of
                # zeros in order to avoid issues with rays hitting the exact
                # location of a node
                normal = numpy.array([(i + 1)*1.23456789e-8 for i
                                      in range(self.ndims)])
                normal[axis] = pm

                distance = pls*(self.xmaxs[axis] - point[axis]) + \
                    mns*(point[axis] - self.xmins[axis])
                if distance < minDistance:
                    self.direction = normal * max(distance, minDistance)
                    minDistance = distance

        for i in range(self.ndims):
            p, d = point[i], self.direction[i]
            self.xminRay[i] = min(p, p + d)
            self.xmaxRay[i] = max(p, p + d)

    def computeIntersection(self, point, poly):
        """
        Compute the intersection with either a triangle or a quadrilateral
        @param point starting point of the ray
        @param poly either a triangle or quad
        @return lmbda (linear parametric coordinate), xis (triangle parametric coordinates)
        """
        ip0 = poly[0]
        self.b = self.points[ip0] - point
        self.mat[:, 1] = self.points[ip0] - self.points[poly[1]]
        self.mat[:, 2] = self.points[ip0] - self.points[poly[-1]]
        solution = numpy.linalg.solve(self.mat, self.b)
        return solution[0], solution[1:]


##############################################################################
def test():

    from icqsol.shapes.ShapeManager import ShapeManager

    shape_mgr = ShapeManager('vtk', 'POLYDATA')
    shp = shape_mgr.createShape('sphere', origin=(0., 0., 0.),
                                radius=1.0, n_theta=6, n_phi=3)

    inside = Inside(shp)

    pt = numpy.array([0., 0., 0.])
    assert(inside.isInside(pt, 0.) == 1)

    pt = numpy.array([1.01, 0., 0.])
    assert(inside.isInside(pt, 0.) == -1)

    pt = numpy.array([0., 0.9999999999, 0.])
    assert(inside.isInside(pt, 0.0) == 1)
    assert(inside.isInside(pt, 0.01) == 0)

    pt = numpy.array([0., 1.0000000001, 0.])
    assert(inside.isInside(pt, 0.) == -1)
    assert(inside.isInside(pt, 0.01) == 0)

if __name__ == '__main__':
    test()



#!/usr/bin/env python

"""
@brief A base class for constructing shapes
@author alexander@gokliya.net
"""

from __future__ import print_function
from csg.core import CSG
from csg.geom import Vector
import numpy

DEFAULTS = dict(origin=[0.0, 0.0, 0.0],
                lengths=[1.0, 1.0, 1.0],
                radius=1.0,
                angle=90.0,
                n_theta=16,
                n_phi=8)

def Box(origin, lengths):
    """
    Create box
    @param  origin/low  end  of  the  box
    @param  lengths  lengths  in  x,  y,  and  z
    """
    center = [origin[i] + 0.5*lengths[i] for i in range(len(origin))]
    radius = [0.5*le for le in lengths]
    return CSG.cube(center=center, radius=radius)


def Cone(radius, origin, lengths, n_theta=16):
    """
    Create cone
    @param radius radius
    @param origin location of the focal point
    @param lengths lengths of the cone
    @param n_theta number of theta cells
    """
    ori = Vector(origin[0], origin[1], origin[2])
    end = Vector(origin[0] + lengths[0],
                 origin[1] + lengths[1],
                 origin[2] + lengths[2])
    return CSG.cone(start=ori,
                    end=end,
                    radius=radius,
                    slices=n_theta)


def Cylinder(radius, origin, lengths, n_theta=16):
    """
    Create cylinder
    @param radius radius
    @param origin center of low end disk
    @param lengths lengths of the cylinder along each axis
    @param n_theta number of theta cells
    """
    ori = Vector(origin[0], origin[1], origin[2])
    end = Vector(origin[0] + lengths[0],
                 origin[1] + lengths[1],
                 origin[2] + lengths[2])
    return CSG.cylinder(start=ori,
                        end=end,
                        radius=radius,
                        slices=n_theta)


def Sphere(radius, origin, n_theta=16, n_phi=8):
    """
    Create sphere
    @param radius radius
    @param origin center of the sphere
    @param n_theta number of theta cells
    @param n_phi number of azimuthal cells
    """
    return CSG.sphere(center=origin,
                      radius=radius,
                      slices=n_theta,
                      stacks=n_phi)


def CompositeShape(shape_tuples=[], expression=''):
    """
    @param shape_tuples list of (variable_name, shape) pairs
    @param expression expression involving +, -, and * operations.
    """
    for i in range(len(shape_tuples)):
        varName = shape_tuples[i][0]
        cmd = '{0} = shape_tuples[{1}][1]'.format(varName, i)
        exec(cmd)
    return eval(expression)



#!/usr/bin/env python
"""
@brief A base class for constructing shapes
"""
from __future__ import print_function
import os
import re
import vtk
import numpy
# We need the following to handle expressions received from callers.
from numpy import linspace
from math import sqrt, sin, cos, tan, log, exp, pi, asin, acos, atan, atan2, e

from csg.geom import Vector, Vertex, Polygon, BSPNode
from csg.core import CSG
from icqsol.shapes.icqShape import Box, Cone, Cylinder, Sphere
from icqsol.shapes.icqShape import DEFAULTS, CompositeShape
from icqsol.color.icqColorMap import ColorMap
from icqsol.shapes.icqRefineSurface import RefineSurface
from icqsol.shapes.icqCoarsenSurface import CoarsenSurface

LOCATIONS = ['POINT', 'CELL']
VTK_DATASET_TYPES = ['STRUCTURED_GRID', 'POLYDATA', 'UNSTRUCTURED_GRID']
FILE_FORMATS = ['ply', 'vtk']


class ShapeManager(object):

    def __init__(self, file_format=None, vtk_dataset_type=None):
        """
        This class incorporates features from 2 primary classes: CSG and
        VTK.  It uses CSG to assemble shapes, which has no knowledge of
        fields.  VTK data objects are grids with fields typically attached
        to them, but not always.

        If file_format is 'vtk', this class's methods work with the POLYDATA
        vtk_dataset_type, so other types are converted as a first step.
        """
        self.file_format = file_format
        self.vtk_dataset_type = vtk_dataset_type
        self.vtk_geometry_filter = None
        if self.file_format is None:
            self.reader = None
            self.writer = None
        elif self.file_format == 'vtk':
            assert self.vtk_dataset_type in VTK_DATASET_TYPES, \
                "Invalid vtk_dataset_type %s" % str(self.vtk_dataset_type)
            self.setVtkGeometryFilter()
            self.setReader(self.file_format, self.vtk_dataset_type)
            self.setWriter(self.file_format, self.vtk_dataset_type)
        else:
            # We have a PLY file
            self.setReader(self.file_format)
            self.setWriter(self.file_format)

    def createShape(self, type, origin=None, lengths=None, radius=None,
                    angle=None, n_theta=None, n_phi=None):
        """
        Create a primitive shape which can be one of box, cone, cylinder or
        sphere.
        @param type the type of shape: box, cone, cylinder or sphere
        @param origin an (x,y,z) tuple consisting of float origin coordinates
        @param lengths (optional) float lengths in the (x,y,z) directions
        @param radius (optional) float radius
        @param angle (optional) float angle
        @param n_theta (optional) number of longitudes (if applicable)
        @param n_phi (optional) number of latitudes (if applicable)
        """
        # Set defaults if necessary.
        if origin is None:
            origin = DEFAULTS['origin']
        if lengths is None:
            lengths = DEFAULTS['lengths']
        if radius is None:
            radius = DEFAULTS['radius']
        if angle is None:
            angle = DEFAULTS['angle']
        if n_theta is None:
            n_theta = DEFAULTS['n_theta']
        if n_phi is None:
            n_phi = DEFAULTS['n_phi']
        # Create the specified shape.
        if type == 'box':
            return Box(origin, lengths)
        if type == 'cone':
            return Cone(radius, origin, lengths, n_theta)
        if type == 'cylinder':
            return Cylinder(radius, origin, lengths, n_theta)
        if type == 'sphere':
            return Sphere(radius, origin, n_theta, n_phi)
        return None

    def addTextureToVtkPolyData(self, vtk_poly_data,
                                texture_file,
                                max_edge_length=float('inf'),
                                texture_file_format=''):
        """
        Add texture to a vtkPolyData object
        @param vtk_poly_data, VTKPolyData object
        @param texture_file texture file (jpg or png)
        @param max_edge_length maximum edge length, refine if need be
        @param texture_file_format texture file format (jpg or png)
        @return new vtkPolyDataObject
        """

        # Select the reader. The user can either explicitly set the format
        # or the reader can be selected implicitly from the file suffix.
        reader = None
        if texture_file_format.lower().find('jp') >= 0 \
             or texture_file.lower().find('.jp') > 0:
                reader = vtk.vtkJPEGReader()
        elif texture_file_format.lower().find('png') >= 0 \
             or texture_file.lower().find('.png') > 0:
                reader = vtk.vtkPNGReader()
        if not reader:
            msg = 'Could not choose reader from file format {0} or file name {1}'
            raise NotImplementedError(
                        msg.format(texture_file_format, texture_file)
                                      )

        reader.SetFileName(texture_file)
        reader.Update()

        imageData = reader.GetOutput()

        # number of pixels
        n0, n1, one = imageData.GetDimensions()

        # box bounds
        xmin, xmax, ymin, ymax, zmin, zmax = vtk_poly_data.GetBounds()

        # mid position of the box
        xmid = (xmin + xmax)/2.
        ymid = (ymin + ymax)/2.
        zmid = (zmin + zmax)/2.
        xlen = xmax - xmin
        ylen = ymax - ymin
        zlen = zmax - zmin
        midPos = numpy.array([xmid, ymid, zmid])
        
        # extrude a little to ensure that the projection box
        # lie beyond the object
        xlen *= 1.01
        ylen *= 1.01
        zlen *= 1.01        
        xmin = xmid - 0.5*xlen
        xmax = xmid + 0.5*xlen
        ymin = ymid - 0.5*ylen
        ymax = ymid + 0.5*ylen
        zmin = zmid - 0.5*zlen
        zmax = zmid + 0.5*zlen

        # face normals of the box
        faceUnitVecs = [numpy.array([+1., 0., 0.]),
                        numpy.array([0., +1., 0.]),
                        numpy.array([0., 0., +1.]),
                        numpy.array([-1., 0., 0.]),
                        numpy.array([0., -1., 0.]),
                        numpy.array([0., 0., -1.])]

        # the index that does not vary (index where the above
        # normals are non-zero)
        constIndex = [0, 1, 2, 0, 1, 2]

        # the parametric u, v unit vector tangential to each face. u cross v
        # gives the normal
        uvVecs = [(numpy.array([0., +1., 0.]), numpy.array([0., 0., +1.])),
                  (numpy.array([-1., 0., 0.]), numpy.array([0., 0., +1.])),
                  (numpy.array([-1., 0., 0.]), numpy.array([0., -1., 0.])),
                  (numpy.array([0., 0., -1.]), numpy.array([0., -1., 0.])),
                  (numpy.array([0., 0., -1.]), numpy.array([+1., 0., 0.])),
                  (numpy.array([0., +1., 0.]), numpy.array([+1., 0., 0.]))]
                  
        uvLengths = [(ylen, zlen),
                     (xlen, zlen),
                     (xlen, ylen),
                     (zlen, ylen),
                     (zlen, xlen),
                     (ylen, xlen)]

        # the low end start points for each face
        startPoints = [numpy.array([xmax, ymin, zmin]),
                       numpy.array([xmax, ymax, zmin]),
                       numpy.array([xmax, ymax, zmax]),
                       numpy.array([xmin, ymax, zmax]),
                       numpy.array([xmin, ymin, zmax]),
                       numpy.array([xmin, ymin, zmin])]

        # d0 is the number of local indices on each tile (along u and v)
        d = min(n0 // 4, n1 // 3)

        def getImageIndices(xyz):
            """
            Map a position on the object to a set of two indices
            which can be used to retrieve the color from the texture
            file
            @param xyz a point
            @return i0, i1 indices, 0 <= i0 < n0 and 0 <= i1 < n1
            """

            # direction of the ray
            direction = xyz - midPos
            # make sure it is not zero
            direction += 1.e-10*numpy.array([1.234567, 2.3456789, 3.45678])

            # find  the face that is most aligned to the ray
            faceIndex = numpy.argmax([fUnit.dot(direction) for
                                      fUnit in faceUnitVecs])

            uVec = uvVecs[faceIndex][0]
            vVec = uvVecs[faceIndex][1]
            startPos = startPoints[faceIndex]

            # find the intersection point of the ray with the face
            ci = constIndex[faceIndex]
            lambd = (startPos[ci] - midPos[ci])/direction[ci]
            projectedPoint = midPos + lambd*direction

            # global tile indices, the tiles are arranged in staircase fashion
            tileI = (faceIndex + 1) // 2
            tileJ = faceIndex // 2

            # local indices on the tile
            j0 = int(d * uVec.dot(projectedPoint - startPos) / uvLengths[faceIndex][0])
            j1 = int(d * vVec.dot(projectedPoint - startPos) / uvLengths[faceIndex][1])

            # the image indices
            i0, i1 = tileI*d + j0, tileJ*d + j1

            return i0, i1

        # Refine if need be.
        pdata = self.refineVtkPolyData(vtk_poly_data,
                                       max_edge_length=max_edge_length)

        # set the colors from the texture file
        rgbArray = vtk.vtkUnsignedCharArray()
        rgbArray.SetNumberOfComponents(3)
        numPoints = pdata.GetNumberOfPoints()
        rgbArray.SetNumberOfTuples(numPoints)
        rgbArray.SetName('Colors')
        point_data_array = imageData.GetPointData().GetArray(0)
        for i in range(numPoints):
            xyz = pdata.GetPoint(i)
            i0, i1 = getImageIndices(xyz)
            rgba = point_data_array.GetTuple(i0 + n0*i1)
            rgbArray.SetTuple(i, rgba[:3])

        pdata.GetPointData().SetScalars(rgbArray)

        return pdata

    def getFieldCentering(self, vtk_poly_data, field_name):
        """
        Get array centering
        @param vtk_poly_data, vtkPolyData instance
        @param field_name name of the field to integrate
        return True if array is nodal, False if cell centered        
        """
        isPoint = False
        array = vtk_poly_data.GetPointData().GetScalars(field_name)
        if array is None:
            array = vtk_poly_data.GetCellData().GetScalars(field_name)
        else:
            isPoint = True
        # Bail out if field was not found
        if array is None:
            raise NotImplementedError(
              'Could not find field "{0}"!'.format(field_name))
        return isPoint

    def getFieldRange(self, vtk_poly_data, field_name,
                      field_component=0):
        """
        Get the min/max field values
        @param vtk_poly_data, vtkPolyData instance
        @param field_name name of the field to integrate
        @param field_component field component
        @return min, max values        
        """
        # Get array and centering.
        isPoint = self.getFieldCentering(vtk_poly_data, field_name)
        array = None
        if isPoint:
            array = vtk_poly_data.GetPointData().GetScalars(field_name)
        else:
            array = vtk_poly_data.GetCellData().GetScalars(field_name)
        numComps = array.GetNumberOfComponents()
        assert field_component < numComps, \
            "Field component {0} must be < {1}".format(field_component, numComps)
        return array.GetRange(field_component)
        
    def integrateSurfaceField(self, vtk_poly_data, field_name,
                              field_component=0):
        """
        Surface integral of a field (point or cell)
        @param vtk_poly_data, vtkPolyData instance
        @param field_name name of the field to integrate
        @param field_component field component
        @return surface integral
        """
        res = 0
        # Determine if the field is point or cell centered.
        isPoint = self.getFieldCentering(vtk_poly_data, field_name)
        if isPoint:
            array = vtk_poly_data.GetPointData().GetScalars(field_name)
        else:
            array = vtk_poly_data.GetCellData().GetScalars(field_name)
        numComps = array.GetNumberOfComponents()
        assert field_component < numComps, \
            "Field component {0} must be < {1}".format(field_component, numComps)
        
        # Iterate over all the polys.
        points = vtk_poly_data.GetPoints()
        cells = vtk_poly_data.GetPolys()
        numCells = cells.GetNumberOfCells()
        ptIds = vtk.vtkIdList()
        cells.InitTraversal()
        if isPoint:
            # Point array. Average the nodal field to get the 
            # cell centered value
            for i in range(numCells):
                cell = cells.GetNextCell(ptIds)
                npts = ptIds.GetNumberOfIds()
                if npts < 3: 
                    continue
                ia = ptIds.GetId(0)
                pa = numpy.array(points.GetPoint(ia))
                fa = array.GetComponent(ia, field_component)
                for j in range(1, npts - 1):
                    ib, ic = ptIds.GetId(j), ptIds.GetId(j + 1)
                    pb = numpy.array(points.GetPoint(ib))
                    pc = numpy.array(points.GetPoint(ic))
                    pb -= pa
                    pc -= pa
                    areaTimesTwo = numpy.linalg.norm(numpy.cross(pb, pc))
                    fb = array.GetComponent(ib, field_component)
                    fc = array.GetComponent(ic, field_component)
                    res += areaTimesTwo * (fa + fb + fc) / 6.0
        else:
            # Cell centered field
            for i in range(numCells):
                cell = cells.GetNextCell(ptIds)
                npts = ptIds.GetNumberOfIds()
                if npts < 3:
                    continue
                ia = ptIds.GetId(0)
                pa = numpy.array(points.GetPoint(ia))
                f = array.GetComponent(i, field_component)
                for j in range(1, npts - 1):
                    ib, ic = ptIds.GetId(j), ptIds.GetId(j + 1)
                    pb = numpy.array(points.GetPoint(ib))
                    pc = numpy.array(points.GetPoint(ic))
                    pb -= pa
                    pc -= pa
                    areaTimesTwo = numpy.linalg.norm(numpy.cross(pb, pc))
                    res += areaTimesTwo * f / 2.0
                    
        return res

    def addSurfaceFieldFromExpressionToVtkPolyData(self, vtk_poly_data,
                                                   field_name,
                                                   expression,
                                                   time_points,
                                                   max_edge_length=float('inf'),
                                                   location='POINT'):
        """
        Add a surface field to a shape using an expression consisting of
        legal variables x,y,z (shape point coordinates) and t (time).
        @param vtk_poly_data, VTKPolyData converted from shape
        @param field_name, the name of the surface field
        @param expression, expression of variables x, y, z, and t
        @param time_points, list of floating point values defining
               snapshots in a time sequence
        @param max_edge_length maximum edge length, refine if need be
        @param location location of field within cell, either 'POINT' or 'CELL'
        @return vtkPolyData instance
        """
        # Sometimes the received expression has been mangled.
        for op in [('__gt__', '>'), ('__lt__', '<'),
                   ('__ge__', '>='), ('__le__', '<='),
                   ('__eq__', '=='), ('__ne__', '!=')]:
            expression = re.sub(op[0], op[1], expression)
        
        # Refine if need be.
        pdata = self.refineVtkPolyData(vtk_poly_data,
                                       max_edge_length=max_edge_length)
        # The field name cannot have spaces.
        valid_field_name = field_name.replace(' ', '_')
        # Get the points from the shape.
        points = pdata.GetPoints()
        num_points = points.GetNumberOfPoints()
        # Define the data.
        data = vtk.vtkDoubleArray()
        data.SetName(valid_field_name)
        # Handle time points.
        num_time_points = len(time_points)
        # Update the data.
        data.SetNumberOfComponents(num_time_points)
        if location.upper() == 'POINT':
            data.SetNumberOfTuples(num_points)
            # Set the surface field values.
            for i in range(num_points):
                x, y, z = points.GetPoint(i)
                for j in range(num_time_points):
                    t = time_points[j]
                    field_value = eval(expression)
                    data.SetComponent(i, j, field_value)
            # Add the field.
            pdata.GetPointData().AddArray(data)
        elif location.upper() == 'CELL':
            num_cells = pdata.GetNumberOfCells()
            data.SetNumberOfTuples(num_cells)
            # Set the surface field values.
            ptIds = vtk.vtkIdList()
            for i in range(num_cells):
                pdata.GetCellPoints(i, ptIds)
                avrgPosition = numpy.zeros((3,), numpy.float64)
                num_pts = ptIds.GetNumberOfIds()
                for k in range(num_pts):
                    avrgPosition += numpy.array(
                        points.GetPoint(ptIds.GetId(k)))
                avrgPosition /= num_pts
                x, y, z = avrgPosition
                for j in range(num_time_points):
                    t = time_points[j]
                    field_value = eval(expression)
                    data.SetComponent(i, j, field_value)
            # Add the field.
            pdata.GetCellData().AddArray(data)

        return pdata

    def addSurfaceFieldFromExpressionToShape(self, shape, field_name,
                                             expression,
                                             time_points,
                                             max_edge_length=float('inf'),
                                             location='POINT'):
        """
        Add a surface field to a shape using an expression consisting of
        legal variables x,y,z (shape point coordinates) and t (time).
        @param shape
        @param field_name the name of the surface field
        @param expression expression consisting containing variables
                          x, y, z, and t
        @param time_points list of floating point values defining
               snapshots in a time sequence
        @param max_edge_length max edge length for refinement
        @param location location of field within cell, either 'POINT' or 'CELL'
                        (capitalization does not matter)
        @return pdata VTKPolyData converted from shape with added surface field
        """
        # Get the points from the shape.
        pdataInput = self.shapeToVTKPolyData(shape)

        # Refine if need be.
        pdata = self.refineVtkPolyData(pdataInput,
                                       max_edge_length=max_edge_length)
        return self.addSurfaceFieldFromExpressionToVtkPolyData(pdata,
                                                               field_name,
                                                               expression,
                                                               time_points,
                                                               max_edge_length,
                                                               location)

    def colorSurfaceField(self, vtk_poly_data, color_map,
                          field_name='', field_component=0):
        """
        Color a selected surface field of a shape.
        @param vtk_poly_data data defining the shape with
                             surface field information
        @param color_map, name of the color map to use
        @param field_name, the name of the surface field to color
        @param field_component field component
        @return colored_vtk_poly_data, color applied to vtk_poly_data
        """
        # Copy the received vtk_poly_data, creating another
        # vtkPolyData object with the same points and cells.
        vtk_poly_data_copy = vtk.vtkPolyData()
        vtk_poly_data_copy.SetPoints(vtk_poly_data.GetPoints())
        vtk_poly_data_copy.SetPolys(vtk_poly_data.GetPolys())
        # Get information from the data.
        isPoint = self.getFieldCentering(vtk_poly_data, field_name)
        if isPoint:
            array = vtk_poly_data.GetPointData().GetScalars(field_name)
        else:
            array = vtk_poly_data.GetCellData().GetScalars(field_name)
        numComps = array.GetNumberOfComponents()
        assert field_component < numComps, \
            "Field component should be < {0}".format(numComps)
        # Get the min/max field values.
        fmin, fmax = array.GetRange()
        # Prepare for coloring.
        rgbArray = vtk.vtkUnsignedCharArray()
        rgbArray.SetNumberOfComponents(3)
        numElements = array.GetNumberOfTuples()
        rgbArray.SetNumberOfTuples(numElements)
        rgbArray.SetName('Colors')
        # Build the color map.
        colorMap = ColorMap(fmin, fmax)
        colorMethod = eval('colorMap.' + color_map)
        # Set the colors.
        for i in range(numElements):
            f = array.GetComponent(i, field_component)
            rgbArray.SetTuple(i, colorMethod(f))
        # Attach the array.
        if isPoint:
            vtk_poly_data_copy.GetPointData().SetScalars(rgbArray)
        else:
            vtk_poly_data_copy.GetCellData().SetScalars(rgbArray)
        return vtk_poly_data_copy

    def cleanPolygon(self, verts, min_triangle_area):
        """
        Remove vertices that are degenerate
        @param verts list of Vertex instances (input and output)
        @param min_triangle_area minimum triangle area
        """
        numVerts = len(verts)
        if numVerts < 3:
            # Nothing to do.
            return
        pa = verts[0].pos
        area = verts[1].pos.minus(pa).cross(verts[2].pos.minus(pa)).length()
        if area > min_triangle_area:
            # We can compute the normal.
            return
        else:
            # Remove vertex and recursively call this method.
            del verts[1]
            self.cleanPolygon(verts, min_triangle_area)

    def refineShape(self, shape, refine=1):
        """
        Refine a shape by inserting one vertex to each edge and one
        vertex to the center of the polygon
        @param shape shape
        @param refine number of refinements
        @return new shape
        """
        s = shape.clone()
        for i in range(refine):
            s = s.refine()
        return s

    def refineVtkPolyData(self, polydata, max_edge_length):
        """
        Refine a vtkPolyData object by adding points along cell edges
        @param polydata vtkPolyData instance
        @param max_edge_length maximum edge length, edges smaller than
                               this value will not be segmented
        @return vtkPolyData instance
        """
        rs = RefineSurface(polydata)
        rs.refine(max_edge_length=max_edge_length)
        return rs.getVtkPolyData()

    def coarsenVtkPolyData(self, polydata, min_cell_area):
        """
        Coarsen a vtkPolyData object by coalescing cells
        @param polydata vtkPolyData instance
        @param min_cell_area minimum cell area beyond which cells
               will be merged
        @return vtkPolyData instance
        """
        cs = CoarsenSurface(polydata)
        cs.coarsen(min_cell_area=min_cell_area)
        return cs.getVtkPolyData()

    def cloneShape(self, shape):
        """
        Clone shape
        @param shape shape
        @return new shape
        """
        return shape.clone()

    def composeShapes(self, shape_tuples=[], expression=''):
        """
        Compose shapes into a more complex shape.
        @param shape_tuples list of (variable_name, shape) pairs
        @param expression expression involving +, -, and * operations.
        @return new shape
        """
        return CompositeShape(shape_tuples, expression)

    def getBoundarySurfaceInsideShape(self, shape, other):
        """
        Return the portion of the surface that is inside another shape
        @param shape
        @param other other shape
        @return shape
        """
        a = BSPNode(shape.clone().polygons)
        b = BSPNode(other.clone().polygons)
        b.invert()
        a.clipTo(b)
        return CSG.fromPolygons(a.allPolygons())

    def loadAsVtkData(self, file_name):
        """
        Load a subclass of a vtkData object from a file
        where a subclass is one of vtkStructuredPoints(),
        vtkStructuredGrid(), vtkPolyData(), vtkRectilinearGrid()
        or vtkUnstructuredGrid()
        @param file_name suffix should be .ply or .vtk
        @return a subclass of a VtkData object
        """
        if not os.path.exists(file_name):
            raise IOError('File {0} not found'.format(file_name))
        self.reader.SetFileName(file_name)
        self.reader.Update()
        return self.reader.GetOutput()

    def loadAsVtkPolyData(self, file_name):
        """
        Load a vtkPolyData object from a file.
        @param file_name, suffix should be .ply or .vtk
        @return a vtkPolyData object
        """
        vtk_data = self.loadAsVtkData(file_name)
        if self.file_format == 'vtk':
            if self.vtk_dataset_type == 'POLYDATA':
                vtk_poly_data = vtk_data
            else:
                vtk_poly_data = self.convertToPolyData(vtk_data)
        else:
            # We have a PLY file.
            vtk_poly_data = vtk_data
        return vtk_poly_data

    def loadAsShape(self, file_name):
        """
        Load an in-memory shape from a file
        @param file_name suffix should be .ply or .vtk
        @return a Shape object
        """
        vtk_poly_data = self.loadAsVtkPolyData(file_name)
        return self.shapeFromVTKPolyData(vtk_poly_data)

    def rotateVtkPolyData(self, pdata, axis=(1., 0., 0.), angleDeg=0.0):
        """
        Rotate vtkPolyData object along given axis
        @param pdata vtkPolyData instance (modified on output)
        @param axis rotation axis
        @param angleDeg angle in degrees
        """
        transform = vtk.vtkTransform()
        transform.RotateWXYZ(angleDeg, axis[0], axis[1], axis[2])
        transformFilter = vtk.vtkTransformPolyDataFilter()
        transformFilter.SetTransform(transform)
        if vtk.VTK_MAJOR_VERSION >= 6:
            transformFilter.SetInputData(pdata)
        else:
            transformFilter.SetInput(pdata)
        transformFilter.Update()
        pdata.DeepCopy(transformFilter.GetOutput())

    def translateVtkPolyData(self, pdata, displ=(0., 0., 0.)):
        """
        Translate vtkPolyData
        @param pdata vtkPolyData instance (modified on output)
        @param displ displacement vector
        """
        transform = vtk.vtkTransform()
        transform.Translate(displ[0], displ[1], displ[2])
        transformFilter = vtk.vtkTransformPolyDataFilter()
        transformFilter.SetTransform(transform)
        if vtk.VTK_MAJOR_VERSION >= 6:
            transformFilter.SetInputData(pdata)
        else:
            transformFilter.SetInput(pdata)
        transformFilter.Update()
        pdata.DeepCopy(transformFilter.GetOutput())

    def scaleVtkPolyData(self, pdata, factors=(1., 1., 1.)):
        """
        Translate vtkPolyData
        @param pdata vtkPolyData instance (modified on output)
        @param amplification vector
        """
        transform = vtk.vtkTransform()
        transform.Scale(factors[0], factors[1], factors[2])
        transformFilter = vtk.vtkTransformPolyDataFilter()
        transformFilter.SetTransform(transform)
        if vtk.VTK_MAJOR_VERSION >= 6:
            transformFilter.SetInputData(pdata)
        else:
            transformFilter.SetInput(pdata)
        transformFilter.Update()
        pdata.DeepCopy(transformFilter.GetOutput())

    def rotateShape(self, shape, axis=(1., 0., 0.), angleDeg=0.0):
        """
        Rotate along axis
        @param shape
        @param axis rotation axis
        @param angleDeg angle in degrees
        """
        return shape.rotate(axis, angleDeg)

    def saveShape(self, shape, file_name, file_type, normals=True):
        """
        Save the shape to a file
        @param shape for saving
        @param file_name file name
        @param file_type either 'ascii' or 'binary'
        @param normals resolve features (corners) and save
                       normal vectors if True
        """
        vtk_poly_data = self.shapeToVTKPolyData(shape)
        self.saveVtkPolyData(vtk_poly_data, file_name, file_type)

    def saveVtkPolyData(self, vtk_poly_data, file_name,
                        file_type, normals=True):
        """
        Save the vtk_poly_data to a file
        @param vtk_poly_data for saving
        @param file_name file name
        @param file_type either 'ascii' or 'binary'
        @param normals resolve features (corners) and save
                       normal vectors if True
        """
        # compute the vertex normals
        if normals:
            vtk_pdata_save = self.computeVertexNormals(vtk_poly_data)
        else:
            vtk_pdata_save = vtk_poly_data

        self.writer.SetFileName(file_name)
        if file_type.lower() == 'ascii':
            self.writer.SetFileTypeToASCII()
        else:
            self.writer.SetFileTypeToBinary()
        if vtk.VTK_MAJOR_VERSION >= 6:
            self.writer.SetInputData(vtk_pdata_save)
        else:
            self.writer.SetInput(vtk_pdata_save)
        self.writer.Write()
        self.writer.Update()

    def shapeFromPolygons(self, polys):
        return CSG.fromPolygons(polys)

    def shapeToPolygons(self, shape):
        return shape.toPolygons()

    def shapeFromVTKPolyData(self, pdata, min_cell_area=1.e-8):
        """
        Create a shape from a VTK PolyData object
        @param pdata vtkPolyData instance
        @param min_cell_area tolerance for cell areas
        @return shape
        @note field data will get lost
        """
        # Store the cell connectivity as CSG polygons.
        numCells = pdata.GetNumberOfPolys()
        cells = pdata.GetPolys()
        cells.InitTraversal()
        ptIds = vtk.vtkIdList()
        polygons = []
        for i in range(numCells):
            cells.GetNextCell(ptIds)
            npts = ptIds.GetNumberOfIds()
            verts = []
            for j in range(npts):
                pointIndex = ptIds.GetId(j)
                pt = pdata.GetPoint(pointIndex)
                v = Vertex(Vector(pt[0], pt[1], pt[2]))
                verts.append(v)
            self.cleanPolygon(verts, min_cell_area)
            if len(verts) >= 3:
                polygons.append(Polygon(verts))
        # Instantiate the shape.
        return CSG.fromPolygons(polygons)

    def shapeToVTKPolyData(self, shape):
        """
        Convert shape to a VTK polydata object
        """
        verts, polys, count = shape.toVerticesAndPolygons()

        points = vtk.vtkPoints()
        numPoints = len(verts)
        points.SetNumberOfPoints(numPoints)
        for i in range(numPoints):
            points.SetPoint(i, verts[i])

        pdata = vtk.vtkPolyData()
        pdata.SetPoints(points)
        
        # build the connectivity
        numPolys = len(polys)
        pdata.Allocate(numPolys, 1)
        ptIds = vtk.vtkIdList()
        for poly in polys:
            numPolyPts = len(poly)
            ptIds.SetNumberOfIds(numPolyPts)
            for j in range(numPolyPts):
                ptIds.SetId(j, poly[j])
            pdata.InsertNextCell(vtk.VTK_POLYGON, ptIds)

        return pdata

    def computeVertexNormals(self, pdata, min_feature_angle=60.0):
        """
        Compute the vertex normals
        @param pdata vtkPolyData instance
        @param min_feature_angle angle above which a sharp feature
                                 is detected. Additional points
                                 will be added to resolve the feature
        @return new vtkPolyData object
        """
        normalsFilter = vtk.vtkPolyDataNormals()
        normalsFilter.SplittingOn()
        normalsFilter.SetFeatureAngle(min_feature_angle)
        normalsFilter.ComputePointNormalsOn()
        if vtk.VTK_MAJOR_VERSION >= 6:
            normalsFilter.SetInputData(pdata)
        else:
            normalsFilter.SetInput(pdata)
        pdata_new = normalsFilter.GetOutput()
        normalsFilter.Update()
        return pdata_new

    def showShape(self, shape, windowSizeX=600, windowSizeY=400, filename=''):
        """
        Show the boundary surface or write image to file
        @param shape CSG instance
        @param windowSizeX number of pixels in x
        @param windowSizeY number of pixels in y
        @param filename write to a file if this keyword
               is present and a non-empty string
        """
        pdata = self.shapeToVTKPolyData(shape)
        self.showVtkPolyData(pdata, windowSizeX, windowSizeY, filename)

    def showVtkPolyData(self, pdata, windowSizeX=600,
                        windowSizeY=400, filename=''):
        """
        Show the boundary surface or write image to file
        @param pdata vtkPolyData instance
        @param windowSizeX number of pixels in x
        @param windowSizeY number of pixels in y
        @param filename write to a file if this keyword
        is present and a non-empty string
        """
        # Create a rendering window and renderer.
        try:
            ren = vtk.vtkRenderer()
            renWin = vtk.vtkRenderWindow()
            iren = vtk.vtkRenderWindowInteractor()
            camera = vtk.vtkCamera()
            mapper = vtk.vtkPolyDataMapper()
            actor = vtk.vtkActor()
            axes = [vtk.vtkArrowSource(),
                    vtk.vtkArrowSource(),
                    vtk.vtkArrowSource()]
            axesTransf = [vtk.vtkTransform(),
                          vtk.vtkTransform(),
                          vtk.vtkTransform()]
            axesTPD = [vtk.vtkTransformPolyDataFilter(),
                       vtk.vtkTransformPolyDataFilter(),
                       vtk.vtkTransformPolyDataFilter()]
            axesMappers = [vtk.vtkPolyDataMapper(),
                           vtk.vtkPolyDataMapper(),
                           vtk.vtkPolyDataMapper()]
            axesActors = [vtk.vtkActor(), vtk.vtkActor(), vtk.vtkActor()]

            renderLarge = vtk.vtkRenderLargeImage()
        except:
            print('WARNING: Cannot call show method -- likely missing VTK')
            return
        renWin.AddRenderer(ren)
        renWin.SetSize(windowSizeX, windowSizeY)
        # Create a renderwindowinteractor.
        iren.SetRenderWindow(renWin)
        # Camera
        xmin, xmax, ymin, ymax, zmin, zmax = pdata.GetBounds()
        lo = numpy.array([xmin, ymin, zmin])
        hi = numpy.array([xmax, ymax, zmax])
        camera.SetFocalPoint(hi)
        center = 0.5*(lo + hi)
        camera.SetPosition(center + hi - lo)
        camera.Zoom(1.0)
        ren.SetActiveCamera(camera)
        # Mapper.
        if vtk.VTK_MAJOR_VERSION >= 6:
            mapper.SetInputData(pdata)
        else:
            mapper.SetInput(pdata)
        # Actor.
        actor.SetMapper(mapper)
        actor.GetProperty().SetColor(1, 1, 1)
        # Add axes.
        axesColrs = [(1., 0., 0.,), (0., 1., 0.,), (0., 0., 1.,)]
        for a in axes:
            a.SetShaftRadius(0.01)
            a.SetTipLength(0.2)
            a.SetTipRadius(0.03)
        for at in axesTransf:
            at.PostMultiply()
        # Rotate the y and z arrows (initially along x).
        axesTransf[1].RotateZ(90.0)
        axesTransf[2].RotateY(-90.0)
        # Scale.
        for i in range(3):
            factor = hi[i] - lo[i]
            scale = [1., 1., 1.]
            scale[i] = factor
            axesTransf[i].Scale(scale)
        # Translate to loBounds.
        for at in axesTransf:
            at.Translate(lo)
        for i in range(3):
            axesTPD[i].SetInputConnection(axes[i].GetOutputPort())
            axesTPD[i].SetTransform(axesTransf[i])
            axesMappers[i].SetInputConnection(axesTPD[i].GetOutputPort())
            axesActors[i].SetMapper(axesMappers[i])
            axesActors[i].GetProperty().SetColor(axesColrs[i])
            ren.AddActor(axesActors[i])
        # Assign actor to the renderer.
        ren.AddActor(actor)
        # Write to file.
        writer = None
        if filename:
            if filename.lower().find('.png') > 0:
                writer = vtk.vtkPNGWriter()
            elif filename.lower().find('.jp') > 0:
                writer = vtk.vtkJPEGWriter()
            elif filename.lower().find('.tiff') > 0:
                writer = vtk.vtkTIFFWriter()
            if writer:
                renderLarge.SetInput(ren)
                renderLarge.SetMagnification(1)
                renderLarge.Update()
                writer.SetFileName(filename)
                writer.SetInputConnection(renderLarge.GetOutputPort())
                writer.Write()
        else:
            # Fire up interactor.
            iren.Initialize()
            renWin.Render()
            iren.Start()

    def translateShape(self, shape, disp=(0., 0., 0.)):
        """
        Translate
        @param shape
        @param disp displacement
        """
        shape.translate(disp)

    def convertToPolyData(self, vtk_data):
        """
        Convert vtk_data to vtk.vtkPolyData().
        @param vtk_data data whose vtk_dataset_type is one of
                         the VTK_DATASET_TYPES besides POLYDATA
        """
        if vtk.VTK_MAJOR_VERSION >= 6:
            self.vtk_geometry_filter.SetInputData(vtk_data)
        else:
            self.vtk_geometry_filter.SetInput(vtk_data)
        self.vtk_geometry_filter.Update()
        vtk_poly_data = self.vtk_geometry_filter.GetOutput()
        return vtk_poly_data

    def getVtkGeometryFilter(self):
        return self.vtk_geometry_filter

    def setVtkGeometryFilter(self):
        """
        Set the vtk_geometry_filter.  We coerce STRUCTURED_GRID and
        UNSTRUCTURED_GRID vtk_dataset_types to be POLYDATA, so we only
        need a vtk.vtkGeometryFilter().
        """
        self.vtk_geometry_filter = vtk.vtkGeometryFilter()

    def chooseReader(self, file_format, vtk_dataset_type=None):
        """
        Return a reader based on file_format and possibly vtk_dataset_type.
        @param vtk_dataset_type, None or one of the VTK_DATASET_TYPES
        """
        # Handle .ply files.
        if file_format == 'ply':
            return vtk.vtkPLYReader()
        # Handle .vtk files.
        if vtk_dataset_type == 'STRUCTURED_GRID':
            return vtk.vtkStructuredGridReader()
        elif vtk_dataset_type == 'POLYDATA':
            return vtk.vtkPolyDataReader()
        elif vtk_dataset_type == 'UNSTRUCTURED_GRID':
            return vtk.vtkUnstructuredGridReader()

    def getReader(self):
        return self.reader

    def setReader(self, file_format=None, vtk_dataset_type=None):
        """
        Set the reader.
        @param file_format, one of vtk or ply - default to self.file_format
        @param vtk_dataset_type, one of the VTK_VTK_DATASET_TYPES
        """
        if file_format is None:
            file_format = self.file_format
        assert file_format in FILE_FORMATS, \
            "Invalid file_format %s" % str(file_format)
        if file_format == 'vtk':
            if vtk_dataset_type is None:
                vtk_dataset_type = self.vtk_dataset_type
            assert vtk_dataset_type in VTK_DATASET_TYPES, \
                "Invalid vtk_dataset_type %s" % str(vtk_dataset_type)
        self.reader = self.chooseReader(file_format, vtk_dataset_type)

    def chooseWriter(self, file_format, vtk_dataset_type):
        """
        Return a writer based on file_format and possibly vtk_dataset_type.
        @param vtk_dataset_type, None or one of the VTK_DATASET_TYPES
        """
        if file_format == 'ply':
            return vtk.vtkPLYWriter()
        # For now we'll just return the POLYDATA writer since methods work
        # only with that vtk_dataset_type.
        return vtk.vtkPolyDataWriter()
        if vtk_dataset_type == 'STRUCTURED_GRID':
            return vtk.vtkStructuredGridWriter()
        elif vtk_dataset_type == 'POLYDATA':
            return vtk.vtkPolyDataWriter()
        elif vtk_dataset_type == 'UNSTRUCTURED_GRID':
            return vtk.vtkUnstructuredGridWriter()

    def getWriter(self):
        return self.writer

    def setWriter(self, file_format=None, vtk_dataset_type=None):
        """
        Set the writer.
        @param file_format, one of vtk or ply - defaults to self.file_format
        @param vtk_dataset_type, None or one of the VTK_VTK_DATASET_TYPES
        """
        if file_format is None:
            file_format = self.file_format
        assert file_format in FILE_FORMATS, \
            "Invalid file_format %s" % str(file_format)
        if file_format == 'vtk':
            if vtk_dataset_type is None:
                vtk_dataset_type = self.vtk_dataset_type
            assert vtk_dataset_type in VTK_DATASET_TYPES, \
                "Invalid vtk_dataset_type %s" % str(vtk_dataset_type)
        self.writer = self.chooseWriter(file_format, vtk_dataset_type)


###############################################################################
def testPrimitiveShapes():
    shape_mgr = ShapeManager(file_format='vtk', vtk_dataset_type='POLYDATA')

    # Box
    box = shape_mgr.createShape('box', origin=(0.,  0.,  0.),
                                lengths=(0.5,  1.,  2.))
    shape_mgr.saveShape(shape=box, file_name='box.vtk', file_type='ascii')
    shape_mgr.showShape(box)

    # Cone
    con = shape_mgr.createShape('cone', radius=1.0, origin=(0.,  0.,  0.),
                                lengths=[1., 0., 0.], n_theta=8)
    shape_mgr.saveShape(shape=con, file_name='con.vtk', file_type='ascii')
    shape_mgr.showShape(con)

    # Cylinder
    cyl = shape_mgr.createShape('cylinder', radius=1.0, origin=(0., 0., 0.),
                                lengths=(1., 0., 0.), n_theta=8)
    shape_mgr.saveShape(shape=cyl, file_name='cyl.vtk', file_type='ascii')
    shape_mgr.showShape(cyl)

    # Sphere
    sph = shape_mgr.createShape('sphere', radius=1.0, origin=(0., 0., 0.),
                                n_theta=8, n_phi=4)
    shape_mgr.saveShape(shape=sph, file_name='sph.vtk', file_type='ascii')
    shape_mgr.showShape(sph)


def testSaveLoad():
    shape_mgr = ShapeManager(file_format='vtk', vtk_dataset_type='POLYDATA')
    s = shape_mgr.createShape('sphere', radius=0.7, origin=(0., 0., 0.),
                              n_theta=8, n_phi=4)
    shape_mgr.saveShape(shape=s, file_name='t.vtk', file_type='ascii')
    s2 = shape_mgr.loadAsShape('t.vtk')
    s3 = shape_mgr.createShape('box', origin=(0.1, 0.2, 0.3),
                               lengths=(1.1, 1.2, 1.3))
    s4 = s2 + s3


def testConstructiveGeometry():
    shape_mgr = ShapeManager(file_format='ply')
    s1 = shape_mgr.createShape('sphere', radius=0.7, origin=(0., 0., 0.))
    s2 = shape_mgr.createShape('sphere', radius=0.2, origin=(0.1, 0.2, 0.3))
    b = shape_mgr.createShape('box', origin=(0.1, 0.2, 0.3),
                              lengths=(1.1, 1.2, 1.3))
    c = shape_mgr.createShape('cylinder', radius=0.5, origin=(0.3, 0.4, 0.5),
                              lengths=(1.0, 0.0, 0.0))
    geom = c*b - s2 - s1
    shape_mgr.saveShape(shape=geom, file_name='geom.ply', file_type='binary')
    geom2 = shape_mgr.loadAsShape('geom.ply')
    shape_mgr.showShape(geom2)


def testShapeComposition():
    shape_mgr = ShapeManager(file_format='vtk', vtk_dataset_type='POLYDATA')
    s1 = shape_mgr.createShape('sphere', radius=1, origin=(0., 0., 0.))
    s2 = shape_mgr.createShape('sphere', radius=1.2, origin=(0.8, 0., 0.))
    s3 = shape_mgr.composeShapes([('s1', s1), ('s2', s2)], 's1 + s2')
    shape_mgr.saveShape(shape=s3, file_name='s3.vtk', file_type='ascii')


if __name__ == '__main__':
    testPrimitiveShapes()
    testSaveLoad()
    testConstructiveGeometry()
    testShapeComposition()



import os
import sys

ASCII = 'ascii'
BINARY = 'binary'

PLY_FORMAT = 'ply'
VTK_FORMAT = 'vtk'

POLYDATA = 'POLYDATA'
STRUCTURED_GRID = 'STRUCTURED_GRID'
UNSTRUCTURED_GRID = 'UNSTRUCTURED_GRID'
VTK_DATASET_TYPES = [POLYDATA, STRUCTURED_GRID, UNSTRUCTURED_GRID]

def getFileFormat(file_name):
    """
    Return a file format based on a file name extension.
    """
    if file_name.lower().endswith('.ply'):
        return PLY_FORMAT
    elif file_name.lower().endswith('.vtk'):
        return VTK_FORMAT

def getVtkDatasetType(file_name):
    """
    Return the VTK dataset type of a file or None if not found.
    """
    # Open the file in binary mode (works for both ASCII and binary files)
    with open(file_name, 'rb') as fh:
        line = fh.readline()
        lineNumber = 1
        # First line: # vtk DataFile Version 3.0
        if line.find(b'vtk') < 0:
            # not a vtk file
            return None
        # Read the first 4 lines
        while lineNumber < 4:
            line = fh.readline()
            lineNumber += 1
        if line.startswith(b'DATASET'):
                return line.split()[1].decode('utf-8')
    return None

def isVtkFile(file_name):
    return file_name.lower().endswith('.vtk')



#!/usr/bin/env python
"""
Analytic expressions for Laplacian singular kernel integrals
http://arxiv.org/pdf/1201.4938.pdf
"""

from __future__ import print_function
from math import cos, sin
from icqsol.bem.icqReferenceTriangle import ReferenceTriangle
from icqsol.bem.icqQuadrature1D import lineQuadrature


class PotentialIntegrals:

    def __init__(self, xa, xb, xc, order=5):
        """
        Constructor
        @param xa first triangle vertex
        @param xb second triangle vertex
        @param xc third triangle vertex
        @param order Gauss integration order
        """
        rt = ReferenceTriangle(xa, xb, xc)
        self.r1 = rt.getR1()
        self.r2 = rt.getR2()
        self.a = rt.getA()
        self.bigTheta = rt.getBigTheta()
        self.sinBigTheta = sin(self.bigTheta)
        self.tanBigTheta = self.sinBigTheta / cos(self.bigTheta)
        self.xbdiff = rt.getPointBDiff()
        self.xcdiff = rt.getPointCDiff()
        self.order = order

    def bigR(self, t):
        return self.r1/(cos(t) - self.a*sin(t))

    def getIntegralOneOverR(self):
        def integrand(t):
            return self.bigR(t)
        return lineQuadrature(self.order, 0.0, self.bigTheta, integrand)

##########################################################################


def testRightTriangle():
    for order in range(1, 6):
        potInt = PotentialIntegrals([0., 0., 0.],
                                    [1., 0., 0.], [0., 1., 0.], order)
        print('order = ', order,
            ' integral of 1/R is ', potInt.getIntegralOneOverR())

if __name__ == '__main__':
    testRightTriangle()



#!/usr/bin/env python

from __future__ import print_function
import vtk
import numpy
from icqsol.shapes.icqRefineSurface import RefineSurface
from icqsol.util.icqDataFetcher import getArrayIndexFromNameAndProjectOntoCells

class BaseSolver:

    def __init__(self, pdata, max_edge_length, order=5):
        """
        Constructor
        @param pdata instance of vtkPolyData
        @param max_edge_length maximum edge length, used to turn
                               polygons into triangles
        """

        # triangulate
        rs = RefineSurface(pdata)
        rs.refine(max_edge_length=max_edge_length)
        self.pdata = rs.getVtkPolyData()

        # store the point indices for each cell
        self.ptIdList = []
        ptIds = vtk.vtkIdList()
        polys = self.pdata.GetPolys()
        polys.InitTraversal()
        for i in range(polys.GetNumberOfCells()):
            polys.GetNextCell(ptIds)
            assert(ptIds.GetNumberOfIds() == 3)
            self.ptIdList.append([ptIds.GetId(0),
                                  ptIds.GetId(1),
                                  ptIds.GetId(2)])

        self.points = self.pdata.GetPoints()
        self.polys = self.pdata.GetPolys()
        self.numTriangles = self.polys.GetNumberOfCells()

        # order of the integration, method dependent
        self.order = order

        # set in the derived classes
        self.responseName = 'NO-SET'
        self.sourceName = 'NOT-SET'

    def getVtkPolyData(self):
        """
        Get the (modified) vtkPolyData object
        @return object
        """
        return self.pdata

    def getPoints(self):
        """
        Get grid points
        @return points
        """
        points = self.pdata.GetPoints()
        numPoints = points.GetNumberOfPoints()
        res = numpy.zeros((numPoints, 3), numpy.float64)
        for i in range(numPoints):
            res[i, :] = points.GetPoint(i)
        return res

    def getCells(self):
        """
        Get cell connectivity
        @return cells
        """
        polys = self.pdata.GetPolys()
        numCells = polys.GetNumberOfCells()
        res = numpy.zeros((numCells, 3), numpy.int)
        polys.InitTraversal()
        ptIds = vtk.vtkIdList()
        for i in range(numCells):
            polys.GetNextCell(ptIds)
            res[i, :] = ptIds.GetId(0), ptIds.GetId(1), ptIds.GetId(2)
        return res
    
    def setResponseFieldName(self, name):
        """
        Set the name of the response field
        @param name name
        """
        self.responseName = name

    def setSourceFieldName(self, name):
        """
        Set the name of the source field
        @param name name
        """
        self.sourceName = name

    def getSourceArrayIndex(self):
        """
        Get the source field index, projecting onto triangles is need be
        """
        srcIndex = getArrayIndexFromNameAndProjectOntoCells(self.pdata, self.sourceName)
        if srcIndex < 0:
            msg = 'ERROR: could not find any cell field named {0}!'.format(self.sourceName)
            raise RuntimeError(msg)
        return srcIndex

    def getSourceArray(self, srcIndex):
        """
        Set the source array 
        @param srcIndex source array index
        """
        srcArray = self.pdata.GetCellData().GetArray(srcIndex)
        n = self.pdata.GetNumberOfPolys()
        src = numpy.zeros((n,), numpy.float64)
        for i in range(n):
            src[i] = srcArray.GetComponent(i, 0)
        return src

    def addResponseField(self, rsp):
        """
        Add the response field to the polydata
        @param rsp response numpy array
        """
        rspData = vtk.vtkDoubleArray()
        rspData.SetNumberOfComponents(1)
        n = len(rsp)
        rspData.SetNumberOfTuples(n)
        rspData.SetName(self.responseName)
        for i in range(n):
            rspData.SetTuple(i, [rsp[i]])
        self.pdata.GetCellData().AddArray(rspData)

    def setSourceFromExpression(self, expression):
        """
        Set the source from expression
        @param expression expression of x, y, and z
        """
        from math import sqrt, pi, sin, cos, tan, log, exp
        
        n = self.pdata.GetNumberOfPolys()
        sourceData = vtk.vtkDoubleArray()
        sourceData.SetNumberOfComponents(1)
        sourceData.SetNumberOfTuples(n)
        sourceData.SetName(self.sourceName)
        midPoint = numpy.zeros((3,), numpy.float64)
        ptIds = vtk.vtkIdList()
        cells = self.pdata.GetPolys()
        cells.InitTraversal()
        for i in range(n):
            cell = cells.GetNextCell(ptIds)
            npts = ptIds.GetNumberOfIds()
            midPoint *= 0 # reset
            for j in range(npts):
                midPoint += self.points.GetPoint(ptIds.GetId(j))
            midPoint /= float(npts)
            x, y, z = midPoint
            v = eval(expression)
            sourceData.SetTuple(i, [v])
        self.pdata.GetCellData().AddArray(sourceData)




#!/usr/bin/env python

from __future__ import print_function
import vtk
import numpy
from icqsol.bem.icqBaseSolver import BaseSolver
from icqsol.bem.icqPotentialIntegrals import PotentialIntegrals
from icqsol.bem.icqQuadrature import gaussPtsAndWeights
from icqsol.util.icqSharedLibraryUtils import getSharedLibraryName
from ctypes import cdll, POINTER, byref, c_void_p, c_double, c_long

FOUR_PI = 4. * numpy.pi

class BaseLaplaceSolver(BaseSolver):

    def __init__(self, pdata, max_edge_length, order=5):
        """
        Constructor
        @param pdata instance of vtkPolyData
        @param max_edge_length maximum edge length, used to turn
                               polygons into triangles
        """

        BaseSolver.__init__(self, pdata, max_edge_length, order)

        libName = getSharedLibraryName('icqLaplaceMatricesCpp')
        self.lib = cdll.LoadLibrary(libName)

        shp = (self.numTriangles, self.numTriangles)
        self.gMat = numpy.zeros(shp, numpy.float64)

        self.__computeResponseMatrix()

    def __computeDiagonalTerms(self):

        # Gauss points and weights
        gpws = gaussPtsAndWeights[self.order]
        npts = gpws.shape[1]
        xsis, etas, weights = gpws[0, :], gpws[1, :], gpws[2, :]

        # iterate over the source triangles
        for jSrc in range(self.numTriangles):

            ia, ib, ic = self.ptIdList[jSrc]

            # The triangle vertex positions
            paSrc = numpy.array(self.points.GetPoint(ia))
            pbSrc = numpy.array(self.points.GetPoint(ib))
            pcSrc = numpy.array(self.points.GetPoint(ic))
            dbSrc = pbSrc - paSrc
            dcSrc = pcSrc - paSrc

            # Iterate over the observer points
            g = 0
            for ipt in range(npts):
                # Observer point
                xObs = paSrc + xsis[ipt]*dbSrc + etas[ipt]*dcSrc
                # Three triangles having observer point as one corner
                pot0ab = PotentialIntegrals(xObs, paSrc, pbSrc, self.order)
                pot0bc = PotentialIntegrals(xObs, pbSrc, pcSrc, self.order)
                pot0ca = PotentialIntegrals(xObs, pcSrc, paSrc, self.order)
                g += weights[ipt] * (pot0ab.getIntegralOneOverR() + \
                                     pot0bc.getIntegralOneOverR() + \
                                     pot0ca.getIntegralOneOverR())

            self.gMat[jSrc, jSrc] = g / (-FOUR_PI)

    def __computeOffDiagonalTerms(self):

        addr = int(self.pdata.GetAddressAsString('vtkPolyData')[5:], 0)
        self.lib.computeOffDiagonalTerms(c_long(addr),
                                         self.gMat.ctypes.data_as(POINTER(c_double)))

    def __computeResponseMatrix(self):

        self.__computeDiagonalTerms()
        self.__computeOffDiagonalTerms()

    def getGreenMatrix(self):
        """
        Return the Green function matrix
        @return matrix
        """
        return self.gMat



from __future__ import print_function
import numpy


class ConjugateGradient:

    def __init__(self, mat, b):
        """
        Constructor
        @param mat dense, square matrix
        @param b right hand side vector
        """
        self.mat = mat
        self.b = b
        n = len(b)
        self.maxNumIters = n
        self.tol = 1.e-10
        self.verbose = False
        self.precond = numpy.array([mat[i, i] for i in range(n)])

    def setTolerance(self, tol):
        """
        Set tolerance
        @param tol tolerance
        """
        self.tol = tol

    def setMaxNumberOfIterations(self, maxNumIters):
        """
        Set maximum number of iterations
        @param maxNumIters  number of iterations
        """
        self.maxNumIters = maxNumIters

    def setVerbosity(self, verbose):
        """
        Set verbosity
        @param verbose True will print(out messages)
        """
        self.verbose = verbose

    def setDiagonalPreconditioner(self, precond):
        """
        Set the diagonal preconditioner
        @param precond preconditioner
        """
        self.precond = precond

    def solve(self, x0):
        """
        Solve linear system
        @param x0 initial guess for solution
        @return solution, error, and number of iterations
        """

        x = x0.copy()
        r = self.b - self.mat.dot(x)
        w = r / self.precond
        p = numpy.zeros(x0.shape, numpy.float64)
        beta = 0.0
        rho = r.dot(w)
        err = numpy.linalg.norm(r)
        k = 0
        while abs(err) > self.tol and k < self.maxNumIters:
            p = w + beta*p
            z = self.mat.dot(p)
            alpha = rho / p.dot(z)
            r -= alpha*z
            w = r / self.precond
            rhoOld = rho
            rho = r.dot(w)
            x += alpha*p
            beta = rho/rhoOld
            err = numpy.linalg.norm(self.b - self.mat.dot(x))
            if self.verbose:
                print('iteration {0} error = {1}'.format(k, err))
            k += 1

        return x, err, k

    def getSolutionError(self, vec):
        """
        Get the solution error
        @param vec solution
        @return error
        """
        return numpy.linalg.norm(self.b - self.mat.dot(vec))



#!/usr/bin/env python

from __future__ import print_function
import math


class ColorMap:


    def __init__(self, fmin, fmax):
        """
        Constructor 
        @param fmin minimum field value
        @param fmax maximum field value
        """

        self.fmin = fmin
        self.fmax = fmax

    @classmethod
    def hatFunc(self, x, x0, x1, x2, x3):
        """
        Hat function
        @param x normalized abscissa, between 0 and 1
        @param x0 x value where function rises from zero
        @param x1 x value where function reaches 1
        @param x2 x value where function drops from 1
        @param x3 x value where function reaches 0
        @return value
        """
        slope01 = 1./(x1 - x0)
        slope23 = 1./(x3 - x2)
        return min(1., max(0., slope01*(x - x0))) \
             - min(1., max(0., slope23*(x - x2)))

    def hot(self, f):
        """
        Get hot color 
        @param f field value
        @return red, green, blue components in range 0 to 255
        """
        x = (f - self.fmin)/(self.fmax - self.fmin)
        r = int(255*math.cos((1. - x)*math.pi/2.0)**2 + 0.5)
        g = int(255*math.sin(x*math.pi)**2 + 0.5)
        b = int(255*math.cos(x*math.pi/2.0)**2 + 0.5)
        return r, g, b

    def cold(self, f):
        """
        Get cold color 
        @param f field value
        @return red, green, blue components in range 0 to 255
        """
        r, g, b = self.hot(f)
        # reverse order
        return b, g, r

    def gnu(self, f):
        """
        Get gnu color
        @param f field value
        @return red, green, blue components in range 0 to 255
        """
        x = (f - self.fmin)/(self.fmax - self.fmin)
        r = int(255*self.hatFunc(x, 3./8., 5./8., 7./8., 9./8.) + 0.5)
        g = int(255*self.hatFunc(x, 1./8., 3./8., 5./8., 7./8.) + 0.5)
        b = int(255*self.hatFunc(x, -1./8., 1./8., 3./8., 5./8.) + 0.5)
        return r, g, b

    def blackbody(self, f):
        """
        Get black body color 
        @param f field value
        @return red, green, blue components in range 0 to 255
        """
        x = (f - self.fmin)/(self.fmax - self.fmin)
        r = int(255*min(1., 2*math.cos(x*math.pi/2.)**2) + 0.5)
        g = int(255*math.sin(x*math.pi)**2 + 0.5)
        b = int(255*min(1., 2*math.cos((1. - x)*math.pi/2.)**2) + 0.5)
        return r, g, b

##############################################################################


def testHat():
    n = 10
    dx = 1.0/float(10)
    x0, x1, x2, x3 = 0.1, 0.3, 0.4, 0.8
    for i in range(n + 1):
        x = i * dx
        print('x = {} y = {}'.format(x, ColorMap.hatFunc(x, x0, x1, x2, x3)))

def testGnu():
    fmin, fmax = 0., 1.
    cm = ColorMap(fmin, fmax)
    n = 10
    df = (fmax - fmin)/float(10)
    for i in range(n + 1):
        f = fmin + i * df
        print('f = {} rgb = {}'.format(f, cm.gnu(f)))

if __name__ == '__main__': 
    testHat()
    testGnu()



#!/usr/bin/env python

from __future__ import print_function
from operator import itemgetter
import math
import numpy
import vtk
import math

class CoarsenSurface:

    EPS = 1.23456789e-10
    TWOPI = 2.0 * math.pi

    def __init__(self, pdata):
        """
        Constructor
        @param pdata vtkPolyData instance
        """
        self.polydata = vtk.vtkPolyData()
        self.polydata.DeepCopy(pdata)

        # will need to be able to interpolate the nodal data to the
        # new vertices
        self.pointData = self.polydata.GetPointData()
        self.numPointData = self.pointData.GetNumberOfArrays()

        # old point id -> new point id
        self.degeneratePtIdMap = {}

        # required so we can get the connectivity between points and 
        # cells
        self.polydata.BuildLinks()
        self.polydata.BuildCells()

    def getVtkPolyData(self):
        """
        Get the vtkPolyData instance associated with the grid
        @return vtkPolyData instance
        """
        return self.polydata

    def getPolygonVectorArea(self, ptIds):
        """
        Get the polygon vector area 
        @param ptIds boundary points of the cell
        @return vector
        """
        points = self.polydata.GetPoints()
        areaVec = numpy.zeros((3,), numpy.float64)
        p0 = numpy.array(points.GetPoint(ptIds.GetId(0)))
        numPts = ptIds.GetNumberOfIds()
        for i in range(1, numPts - 1):
            p1 = numpy.array(points.GetPoint(ptIds.GetId(i    )))
            p2 = numpy.array(points.GetPoint(ptIds.GetId(i + 1)))
            areaVec += numpy.cross(p1 - p0, p2 - p0)
        return areaVec

    def getPolygonArea(self, ptIds):
        """
        Compute the (scalar) area of a polygon
        @param ptIds list of point indices
        @return area
        """
        areaVec = self.getPolygonVectorArea(ptIds)
        return numpy.linalg.norm(areaVec)

    def getPolygonNormalVector(self, ptIds):
        """
        Compute the normal vector of the polygon
        @param ptIds list of point indices
        @return area
        """
        areaVec = self.getPolygonVectorArea(ptIds)
        area = numpy.linalg.norm(areaVec)
        return areaVec / area

    def getTotalAngle(self, ptId):
        """
        Compute the sum of the angles between this point and the surrounding edges
        @param ptId point Id
        @return total angle in radiants, should be about 2*pi for an internal point
        """
        cellIds = vtk.vtkIdList()

        # get the cell Ids sharing this point
        ptIds = vtk.vtkIdList()
        self.polydata.GetPointCells(ptId, cellIds)

        pt0 = numpy.zeros((3,), numpy.float64)
        edge1 = numpy.zeros((3,), numpy.float64)
        edge2 = numpy.zeros((3,), numpy.float64)
        points = self.polydata.GetPoints()

        # get the point coorindates of ptId
        points.GetPoint(ptId, pt0)

        # iterate over the cells sharing ptId
        angle = 0.
        for iCell in range(cellIds.GetNumberOfIds()):

            # get the points of this cell
            self.polydata.GetCellPoints(cellIds.GetId(iCell), ptIds)

            # find the two edges incident to ptId
            n = ptIds.GetNumberOfIds()
            for j in range(n):
                p1 = ptIds.GetId(j)
                p2 = ptIds.GetId((j + 1) % n)
                if p1 == ptId:
                    self.polydata.GetPoints().GetPoint(p2, edge1)
                elif p2 == ptId:
                    self.polydata.GetPoints().GetPoint(p1, edge2)

            # the two edges
            edge1 -= pt0
            edge2 -= pt0

            # add the angle between pt1 and pt2
            crossProduct = numpy.linalg.norm(numpy.cross(edge1, edge2))
            dotProduct = numpy.dot(edge1, edge2)
            angle += math.atan2(crossProduct, dotProduct)

        return angle

    def collapsePolygon(self, cellId):
        """
        Collapse the vertices of a cell
        @param cellId Id of the polygon
        @param return list of points that have been moved
        """
        # points of the cell
        ptIds = vtk.vtkIdList()
        self.polydata.GetCellPoints(cellId, ptIds)
        npts = ptIds.GetNumberOfIds()

        # typically the center of the cell, in some 
        # cases the center of an edge
        center = numpy.zeros( (3,), numpy.float64 )

        # coordinates of the poinrt
        pt = numpy.zeros( (npts,), numpy.float64 )

        points = self.polydata.GetPoints()
        pointsToMove = []
        
        # determine which points are internal/boundary
        internalPointIds = []
        boundaryPointIds = []
        for i in range(npts):
            ptId = ptIds.GetId(i)
            totalAngle = self.getTotalAngle(ptId)
            if abs(totalAngle - self.TWOPI) < 1.e-6: #0.01:
                internalPointIds.append(ptId)
            else:
                boundaryPointIds.append(ptId)
                    
        # compute the (central) point where the cell collapses to
        numBoundaryPoints = len(boundaryPointIds)
        if numBoundaryPoints == 0:
            # all points are internal, average the (internal) points
            for ptId in internalPointIds:
                points.GetPoint(ptId, pt)
                center += pt
            center /= float(len(internalPointIds))
            pointsToMove = internalPointIds
        #elif numBoundaryPoints > 1 and numBoundaryPoints < npts:
        elif numBoundaryPoints > npts and numBoundaryPoints < npts:  # not working well so turning off for the time being
            # average the boundary points and move all points to this position
            for ptId in boundaryPointIds:
                points.GetPoint(ptId, pt)
                center += pt
            center /= float(numBoundaryPoints)
            pointsToMove = internalPointIds + boundaryPointIds
        else:
            # likely all the points are boundary or there are more than two boundary
            # edges, nothing to do
            pass

        # move the selected points to the center
        for ptId in pointsToMove:
            points.SetPoint(ptId, center)

        # store in memory which point ids are degenerate
        for ptId in pointsToMove[1:]:
            self.degeneratePtIdMap[ptId] = pointsToMove[0]

        # average the nodal data at the new vertex location
        self.averagePointData(pointsToMove)

        return pointsToMove

    def coarsen(self, min_cell_area = 1.e-10):
        """
        Coarsen surface by collapsing small polygons
        @param min_cell_area cell area tolerance
        @note operation is in place
        """

        polys = self.polydata.GetPolys()
        numPolys = polys.GetNumberOfCells()
        ptIds = vtk.vtkIdList()
        polys.InitTraversal()
        for cellId in range(numPolys):
            polys.GetNextCell(ptIds)
            area = abs(self.getPolygonArea(ptIds))
            if area < min_cell_area and area > self.EPS:

                # collapse the cell and get the point Ids that have moved
                movedPointIds = self.collapsePolygon(cellId)

                # clean up the connectivity by removing degenerate point Ids
                self.removeDegeneratePointsInConnectivity(movedPointIds)

        self.deleteZeroPolys()

    def removeDegeneratePointsInConnectivity(self, movedPointIds):
        """
        Remove the degenerate point Ids from the connectivity 
        @param movedPointIds list of point Ids that have moved
        """
        neighCellIds = vtk.vtkIdList()
        neighCellPointIds = vtk.vtkIdList()
        if len(movedPointIds) > 1:
            # all other point Ids will be replaced by newPointId
            newPointId = movedPointIds[0]
            for oldPointId in movedPointIds[1:]:
                # get all the cells that contain oldPointId
                self.polydata.GetPointCells(oldPointId, neighCellIds)
                numNeighCells = neighCellIds.GetNumberOfIds()
                for neighCellId in range(numNeighCells):
                    # extract the ptIds of that cell, requires BuildCells 
                    # to be called
                    self.polydata.GetCellPoints(neighCellId, neighCellPointIds)
                    npts = neighCellPointIds.GetNumberOfIds()
                    for i in range(npts):
                        pi = neighCellPointIds.GetId(i)
                        if pi == oldPointId:
                            # replace point Id
                            self.polydata.ReplaceCellPoint(neighCellId, oldPointId, newPointId)

    def deleteZeroPolys(self):
        """
        Delete all the polygons whose areas are (nearly) zero
        """
        ptIds = vtk.vtkIdList()
        numPolys = self.polydata.GetPolys().GetNumberOfCells()
        for polyId in range(numPolys):
            self.polydata.GetCellPoints(polyId, ptIds)
            if abs(self.getPolygonArea(ptIds)) <= self.EPS:
                self.polydata.DeleteCell(polyId)
            
        self.polydata.RemoveDeletedCells()
        self.polydata.BuildLinks() # not sure if this is required
        self.polydata.BuildCells()
 
    def averagePointData(self, pids):
        """
        Average the field at the point locations and set the nodal field values 
        to the average value
        @param pids list of point Ids
        """
        numPts = len(pids)
        if numPts == 0: return

        for el in range(self.numPointData):
            arr = self.pointData.GetArray(el)
            numComps = arr.GetNumberOfComponents()
            vals = numpy.zeros((numComps,), numpy.float64)
            baryVals = numpy.zeros((numComps,), numpy.float64)
            # mid cell values
            for i in range(numPts):
                ptId = pids[i]
                vals[:] = arr.GetTuple(ptId)
                baryVals += vals
            baryVals /= float(numPts)
            # set the field values to the mid cell values
            for j in range(numPts):
                arr.SetTuple(pids[j], baryVals)

##############################################################################


def printVtkPolyData(pdata):

    points = pdata.GetPoints()
    polys = pdata.GetPolys()
    ptIds = vtk.vtkIdList()
    numPolys = polys.GetNumberOfCells()
    print('Number of polygons: {0}'.format(numPolys))
    polys.InitTraversal()
    for i in range(numPolys):
        numPts = ptIds.GetNumberOfIds()
        print('\tCell {0} has {1} points: '.format(i, numPts))
        for j in range(numPts):
            ptId = ptIds.GetId(j)
            pt = points.GetPoint(ptId)
            print('\t\t{0} -> {1}'.format(ptId, pt))


def testNoCoarsening():

    points = vtk.vtkPoints()
    points.InsertNextPoint((0., 0., 0.))
    points.InsertNextPoint((1., 0., 0.))
    points.InsertNextPoint((1., 1., 0.))

    pdata = vtk.vtkPolyData()
    pdata.SetPoints(points)

    ptIds = vtk.vtkIdList()
    ptIds.InsertNextId(0)
    ptIds.InsertNextId(1)
    ptIds.InsertNextId(2)
    pdata.Allocate(1, 1)
    pdata.InsertNextCell(vtk.VTK_POLYGON, ptIds)

    rs = CoarsenSurface(pdata)
    rs.coarsen(min_cell_area=1.e-5)
    pdata2 = rs.getVtkPolyData()
    assert(pdata2.GetNumberOfPolys() == 1)


def testAddingThreePointsThenMore():

    points = vtk.vtkPoints()
    points.InsertNextPoint((0., 0., 0.))
    points.InsertNextPoint((2., 0., 0.))
    points.InsertNextPoint((2., 1., 0.))

    pdata = vtk.vtkPolyData()
    pdata.SetPoints(points)

    ptIds = vtk.vtkIdList()
    ptIds.InsertNextId(0)
    ptIds.InsertNextId(1)
    ptIds.InsertNextId(2)
    pdata.Allocate(1, 1)
    pdata.InsertNextCell(vtk.VTK_POLYGON, ptIds)

    rs = CoarsenSurface(pdata)
    rs.coarsen(min_cell_area=1.e-5)
    assert(rs.getVtkPolyData().GetNumberOfPolys() == 1)


def testStartingWithTwoCells():

    points = vtk.vtkPoints()
    points.InsertNextPoint((0., 0., 0.))
    points.InsertNextPoint((2., 0., 0.))
    points.InsertNextPoint((2., 1., 0.))
    points.InsertNextPoint((0., 1., 0.))

    pdata = vtk.vtkPolyData()
    pdata.SetPoints(points)

    pdata.Allocate(2, 1)
    ptIds = vtk.vtkIdList()
    ptIds.SetNumberOfIds(3)
    ptIds.SetId(0, 0)
    ptIds.SetId(1, 1)
    ptIds.SetId(2, 2)
    pdata.InsertNextCell(vtk.VTK_POLYGON, ptIds)
    ptIds.SetId(0, 2)
    ptIds.SetId(1, 3)
    ptIds.SetId(2, 0)
    pdata.InsertNextCell(vtk.VTK_POLYGON, ptIds)

    rs = CoarsenSurface(pdata)
    rs.coarsen(min_cell_area=1.e-5)
    assert(rs.getVtkPolyData().GetNumberOfPolys() == 2)


if __name__ == '__main__':
    testNoCoarsening()
    testAddingThreePointsThenMore()
    testStartingWithTwoCells()



#!/usr/bin/env python

from __future__ import print_function
import pkg_resources
from ctypes import cdll, POINTER, byref, c_void_p, c_double, c_long
import glob
import sys

PY_MAJOR_VERSION = sys.version_info[0]

# On some distributions the fully qualified shared library name 
# includes suffixes such as '.cpython-35m-darwin.so'
def getFullyQualifiedSharedLibraryName(libName):
    paths = glob.glob(libName + '*')
    # return the first path
    return paths[0]

def getSharedLibraryName(libName):
    if PY_MAJOR_VERSION < 3:
        res = pkg_resources.resource_filename('icqsol', libName + '.so')
    else:
        libName = pkg_resources.resource_filename('icqsol', libName)
        res = getFullyQualifiedSharedLibraryName(libName)
    return res







#!/usr/bin/env python

"""
Set up script for icqsol
alexander@gokliya.net
"""

import os
import glob
from setuptools import setup, Extension
import re
import vtk # for version number
import __init__ # for version number

def anacondaLibraryNameFix(vtkLibraries, vtkLibraryDir):
    """
    On Anaconda, some of the libraries only exist with the vtk version 
    appended.
    """
    vtkVersion = '-{0}.{1}'.format(vtk.VTK_MAJOR_VERSION, vtk.VTK_MINOR_VERSION)
    suffix = '.so'
    if os.uname()[0] == 'Darwin':
        suffix = '.dylib'
    elif os.uname()[0] == 'Windows':
        # I don't have a windows machine to test...
        # do we have to worry about capitalization on Windows?
        suffix = '.dll'
    count = 0
    for lib in vtkLibraries:
        if not os.path.exists(vtkLibraryDir + '/lib' + lib + suffix):
            newName = lib + vtkVersion
            li = vtkLibraryDir + '/lib' + newName + suffix
            if os.path.exists(vtkLibraryDir + '/lib' + newName + suffix) or \
               os.path.exists(vtkLibraryDir + '/lib' + newName + '.a') or \
               os.path.exists(vtkLibraryDir + '/' + newName + suffix):
                # set the loibrary name to name-MAJOR_VERSION.MINOR_VERSION
                vtkLibraries[count] = newName
        count += 1

# Because we're linking C++ code to the VTK library, we 
# need to know where VTK was installed. 
VTK_INCLUDE_DIRS = os.environ.get('VTK_INCLUDE_DIRS', '').split(';')
VTK_LIBRARIES = os.environ.get('VTK_LIBRARIES', '').split(';')
VTK_RUNTIME_LIBRARY_DIRS = os.environ.get('VTK_RUNTIME_LIBRARY_DIRS', '').split(';')
# Override the libraries if one provides VTK_LIBRARY_DIR
vtk_lib_dir = os.environ.get('VTK_LIBRARY_DIR', '')
if vtk_lib_dir:
    libfiles = glob.glob(vtk_lib_dir + '/libvtk*')
    VTK_LIBRARIES = []
    VTK_RUNTIME_LIBRARY_DIRS = [vtk_lib_dir]
    for libfile in libfiles:
        # Skip libraries ending with version number
        if re.search(r'\.\d+$', libfile): continue
        lib = os.path.basename(libfile)
        lib = re.sub(r'^lib', '', lib)
        lib = re.sub(r'\.\w+$', '', lib) 
        # exclude python libraries
        if re.search('Python', lib): continue
        VTK_LIBRARIES.append(lib) 

# Anaconda renamed the libraries
anacondaLibraryNameFix(VTK_LIBRARIES, VTK_RUNTIME_LIBRARY_DIRS[0])

# remove all the Python libraries since they are causing a link issue
# on CentOS 7 and we don't really need them
indicesToDelete = []
for i in range(len(VTK_LIBRARIES)):
  lib = VTK_LIBRARIES[i]
  if lib.find('Python') >= 0 or lib.find('verdict') >=0: 
    indicesToDelete.append(i)
indicesToDelete.reverse()

for i in indicesToDelete:
  del VTK_LIBRARIES[i]

print('VTK_INCLUDE_DIRS         = {0}\n'.format(VTK_INCLUDE_DIRS))
print('VTK_LIBRARIES            = {0}\n'.format(VTK_LIBRARIES))
print('VTK_RUNTIME_LIBRARY_DIRS = {0}\n'.format(VTK_RUNTIME_LIBRARY_DIRS))

setup(name='icqsol',
      version=__init__.__version__, 
      description='Solving engineering problems on the web',
      author='Alex Pletzer and Greg Von Kuster',
      author_email='alexander@gokliya.net',
      url='https://github.com/pletzer/icqsol/wiki',
      install_requires=['pytriangle>=1.0.9', 'pycsg>=0.3.12'],
      dependency_links=['http://github.com/pletzer/pytriangle/tarball/master#egg=pytriangle-1.0.9',
                        'http://github.com/pletzer/pycsg/tarball/master#egg=pycsg-0.3.12',
      ],
      package_dir = {'icqsol': ''}, # the present working directory maps to icqsol below
      data_files = [('icqsol/textures', ['textures/Swietenia_macrophylla_wood.jpg',
                                  'textures/220px-COnglomerate-sandstone_layers_Nerriga.jpg',
                                  'textures/checkerboard.png',])],
      packages=['icqsol',
                'icqsol.color', 
                'icqsol.bem',
                'icqsol.solvers',
                'icqsol.discretization',
                'icqsol.shapes',
                'icqsol.util'],
      ext_modules = [Extension('icqsol.icqLaplaceMatricesCpp', # name of the shared lib
                               ['bem/icqFunctor.cpp',
                                'bem/icqLaplaceFunctor.cpp',
                                'bem/icqQuadrature.cpp',
                                'bem/icqLaplaceMatrices.cpp'] + glob.glob('bem/*.h'),
                               include_dirs=['bem'] + VTK_INCLUDE_DIRS,
                               library_dirs=VTK_RUNTIME_LIBRARY_DIRS,
                               libraries=VTK_LIBRARIES,
                               ),
                      Extension('icqsol.icqInsideLocatorCpp', 
                                ['csg/icqInsideLocator.cpp'] + glob.glob('csg/*.h'),
                                include_dirs=['csg'] + VTK_INCLUDE_DIRS,
                                library_dirs=VTK_RUNTIME_LIBRARY_DIRS,
                                libraries=VTK_LIBRARIES,
                                ),
      ],
      requires = ['numpy', 'vtk',],
     )



#!/usr/bin/env python

from __future__ import print_function
import vtk
import numpy

def getArrayIndexFromName(data, name):
    """
    Get the array index from its name
    @param data either a vtkCellData or a vtkPointData object
    @param name name
    @return index >= 0 if name exists or -1 if it does not
    """
    numArrays = data.GetNumberOfArrays()
    index = -1
    for i in range(numArrays):
        arr = data.GetArray(i)
        if arr.GetName() == name:
            index = i
            break
    return index

def getArrayIndexFromNameAndProjectOntoCells(pdata, name):
    """
    Get array index from its name, if the array 
    is a point data type then project onto cells
    @param pdata polydata instance
    @param name name
    @return index >= 0 if name exists or -1 if it does not
    """
    cellData = pdata.GetCellData()
    pointData = pdata.GetPointData()
    index = getArrayIndexFromName(cellData, name)
    if index < 0:
        # Maybe a point array?
        index2 = getArrayIndexFromName(pointData, name)

        if index2 >= 0:
            # Project from points to cells
            pointArr = pointData.GetArray(index2)
            numCells = pdata.GetNumberOfPolys()
            cellArr = vtk.vtkDoubleArray()
            cellArr.SetName(name) # same name as the point array
            cellArr.SetNumberOfComponents(1)
            cellArr.SetNumberOfTuples(numCells)
            cells = pdata.GetPolys()
            ptIds = vtk.vtkIdList()
            cells.InitTraversal()
            for i in range(numCells):
                cells.GetNextCell(ptIds)
                # We know for sure that all cells are triangles
                ia, ib, ic = ptIds.GetId(0), \
                    ptIds.GetId(1), ptIds.GetId(2)
                va = pointArr.GetComponent(ia, 0)
                vb = pointArr.GetComponent(ib, 0)
                vc = pointArr.GetComponent(ic, 0)
                cellArr.SetComponent(i, 0, (va + vb + vc)/3.)
            # Add the cell array
            cellData.AddArray(cellArr)
            return getPotentialArrayIndexFromName(name)

    return index


